"""See the class docstring."""

from .ArgStringComposer import StructComposer, compose_cmd_string_from_struct_args
from .ArgStructStringParser import ArgStructStringParser
from .StructBase import StructBase
from .InstrumentErrors import RsInstrException


class ArgStructList(object):
	"""Contains methods for composing cmd string and parsing cmd response to the provided structure instance."""

	RAW_DATA_PROP_NAME = 'RawReturnData'

	def __init__(self, struct):
		self.__struct = struct
		self.args = self.__get_struct_arg_attrs()

	def __str__(self):
		return f"'{self.__struct.__class__.__name__}', {len(self.args)} args: [ {', '.join([self.args[x].name for x in self.args])} ]"

	def __get_struct_arg_attrs(self):
		"""Fills self.args with the dictionary of ArgStruct list.
		Dictionary key is the argument order (argument_ix)."""
		# noinspection PyUnresolvedReferences,PyProtectedMember
		args_list = StructBase._StructBase__get_meta_args_list(None, self.__struct)
		# meta_data = getattr(self.__struct, '_{}__meta_args_list'.format(self.__struct.__class__.__name__))
		args = dict()
		for x in args_list:
			args[x.argument_ix] = x

		return args

	def parse_from_cmd_response(self, content: str):
		"""Fills the structure from the entered string content (command response)."""
		if hasattr(self.__struct, ArgStructList.RAW_DATA_PROP_NAME):
			setattr(self.__struct, ArgStructList.RAW_DATA_PROP_NAME, content)

		parser = ArgStructStringParser(self.__struct, content)
		arg_count = len(self.args)
		arg_ix = 0

		# Non-repeated arguments
		while arg_ix < arg_count and self.args[arg_ix].is_open_list is False:
			arg = self.args[arg_ix]
			if arg.repetition <= 1:
				# No list, but scalar value
				parser.to_scalar_value(arg)
			else:
				# List with a fixed size, not repeated
				parser.to_list_value(arg, True, 0, arg.repetition, arg.repetition, 1)

			arg_ix += 1

		if arg_ix < arg_count:
			arg = self.args[arg_ix]
			# Still some args to go
			if arg.is_open_list is True:
				# The previous loop ended because the next argument had is_open_list True
				if arg_ix == (arg_count - 1):
					# This is the last argument, ignore the repetitions and take the whole rest of the elements
					parser.to_list_value(arg, True, 0, parser.remaining, parser.remaining, 1)
				else:
					# More than one argument remaining. Loop through them interleaving the result strings
					open_list_args = {key: value for key, value in self.args.items() if key >= arg_ix}

					# Accumulate the number of repetitions from all the open_list_args
					period: int = sum(open_list_args[ix].repetition for ix in open_list_args)
					reminder: int = parser.remaining % period
					if reminder != 0:
						raise RsInstrException(
							f'Arguments parsing is not aligned - source string elements remaining to parse {parser.remaining}'
							f'is not dividable by the summary Period {period} of all the open list arguments:\n' + '\n'.join(['{}'.format(x) for x in open_list_args]))
					# Go through the arguments and accumulate the list content
					offset = 0
					for x in open_list_args:
						arg = open_list_args[x]
						parser.to_list_value(arg, False, offset, arg.repetition, period, -1)
						offset += arg.repetition

	def compose_cmd_string(self):
		"""Composes the string argument from the structure for sending to the instrument."""
		return compose_cmd_string_from_struct_args(self.args, StructComposer(self.__struct), self.__struct)



"""See the docstring for the IoTransferEventArgs class."""

import itertools
from typing import AnyStr

from .Utilities import size_to_kb_mb_string


class IoTransferEventArgs(object):
	"""Contains event data for driver read or write operations."""
	# first generated is 100
	id_generator = itertools.count(100)

	# noinspection PyTypeChecker
	def __init__(self, reading: bool, opc_sync: bool, total_size: int or None, context: str):
		"""Initializes new instance of IoTransferEventArgs
		:param reading: True: reading operation, False: writing operation
		:param opc_sync: defines if the command is OPC-synchronised
		:param total_size: total size of the data received
		:param context: SCPI query. It is truncated to maximum of 100 characters"""
		self._transfer_id = next(self.id_generator)
		self.end_of_transfer = False
		self.reading = reading
		self.opc_sync = opc_sync

		self.total_size = total_size
		self.transferred_size: int = 0
		self.context: str = (context[:100] + '..') if len(context) > 100 else context

		# Data to set after the object has been created.
		self.chunk_size: int = None
		"""Size of one chunk of data. This number does not change during the transfer."""
		self.chunk_ix: int = None
		"""0-based index of the chunk."""
		self.total_chunks: int = None
		"""Expected number of chunks."""
		self.resource_name: str = None
		"""Visa Resource Name of the instrument that generated the data."""
		self.binary: bool = None
		"""True: Binary data, False: string data"""
		self.data: AnyStr = None
		"""If the feature of transferring data over R/W event is switched ON, this field contains the whole data."""

	@classmethod
	def read_chunk(cls, opc_sync: bool, context: str) -> 'IoTransferEventArgs':
		"""Creates new IoTransferEventArgs of read string \n
		:param opc_sync: defines if the command is OPC-synchronised
		:param context: SCPI query. It is truncated to maximum of 100 characters.
		:return: IoTransferEventArgs object of a read string operation."""
		return cls(True, opc_sync, None, context)

	@classmethod
	def write_str(cls, opc_sync: bool, total_size: int, context: str) -> 'IoTransferEventArgs':
		"""Creates new IoTransferEventArgs of write string \n
		:param opc_sync: defines if the command is OPC-synchronised
		:param total_size: size of the data to write
		:param context: SCPI command write. It is truncated to maximum of 100 characters
		:return: IoTransferEventArgs object of a write-string operation."""
		obj = cls(False, opc_sync, total_size, context)
		obj.binary = False
		return obj

	@classmethod
	def write_bin(cls, context: str) -> 'IoTransferEventArgs':
		"""Creates new IoTransferEventArgs of read binary data \n
		:param context: SCPI command. It is truncated to maximum of 100 characters.
		:return: IoTransferEventArgs object of a write-binary-data operation."""
		# noinspection PyTypeChecker
		obj = cls(False, False, None, context.rstrip())
		obj.binary = True
		return obj

	def __str__(self):
		if self.binary:
			type_info = 'binary'
		else:
			type_info = 'ascii'
		if self.opc_sync:
			type_info += ' (opc-synced)'
		if not self.total_chunks:
			chunk_info = f' chunk nr. {self.chunk_ix + 1}'
		elif self.total_chunks > 1:
			chunk_info = f' chunk nr. {self.chunk_ix + 1}/{self.total_chunks}'
		else:
			chunk_info = ' chunk nr. 1/1'
		eot = ' (EOT)' if self.end_of_transfer else ''
		if self.reading:
			result = f'IoTransferArgs ID {self._transfer_id}: reading {type_info}, {chunk_info} {size_to_kb_mb_string(self.chunk_size, True)}, ' \
					f'sum {size_to_kb_mb_string(self.transferred_size, True)} / {size_to_kb_mb_string(self.total_size, True) if self.total_size else "<N.A.>"}{eot}.'
		else:
			result = f'IoTransferArgs ID {self._transfer_id}: writing {type_info}, {chunk_info} {size_to_kb_mb_string(self.chunk_size, True)}, ' \
						f'sum {size_to_kb_mb_string(self.transferred_size, True)} / {size_to_kb_mb_string(self.total_size, True)}{eot}.'
		if self.context:
			result += f' Cmd: {self.context}'
		return result

	@property
	def transfer_id(self) -> int:
		"""Unique number for each transfer for this Instrument.
		If the transfer is performed in more chunks, the transfer_id stays the same during the whole transfer."""
		return self._transfer_id

	def set_end_of_transfer(self):
		"""Sets fields to signal end of transfer."""
		self.transferred_size = self.total_size
		self.end_of_transfer = True



"""See the class docstring."""

import re
import threading
from enum import Enum
from typing import Callable, Dict
from datetime import datetime, timedelta

from . import Utilities, InstrumentSettings, InstrumentOptions, Conversions as Conv
from .ArgSingleSuppressed import ArgSingleSuppressed
from .ArgStructList import ArgStructList
from .InternalLinker import InternalLinker
from .IoTransferEventArgs import IoTransferEventArgs
from .StreamReader import StreamReader
from .StreamWriter import StreamWriter
from .Utilities import trim_str_response, size_to_kb_mb_string
from .VisaSession import VisaSession, EventArgsChunk
from .VisaSessionSim import VisaSessionSim
from .RepeatedCapability import RepeatedCapability
from .ScpiLogger import ScpiLogger
from .InstrumentErrors import *


class Instrument(object):
	"""Model of an instrument with VISA interface."""

	def __init__(self, resource_name: str, simulate: bool, settings: InstrumentSettings, direct_session=None):
		"""Opening an instrument session.
		If simulate is true, it cannot be later switched to false anymore."""
		self._simulating: bool = simulate
		self._settings = settings
		self._direct_session = direct_session
		self.logger: ScpiLogger or None = None
		self._last_exc_log: str or None = None
		self._start_time: datetime or None = None
		self.__session = None
		self._global_repcaps: Dict[str, RepeatedCapability] = {}
		self._linker = InternalLinker()
		# noinspection PyTypeChecker
		self.on_write_handler: Callable = None
		# noinspection PyTypeChecker
		self.on_read_handler: Callable = None
		self._io_events_include_data: bool = False
		self._lock = None
		self._before_query_handler = None
		self._before_write_handler = None
		# noinspection PyTypeChecker
		self.total_execution_time: timedelta = None
		# noinspection PyTypeChecker
		self.total_time_startpoint: datetime = None
		self.reset_time_statistics()

		if self._settings.selftest_timeout == 0:
			self._settings.selftest_timeout = 100000

		# Changeable settings
		self.resource_name: str = resource_name
		self._instr_options: InstrumentOptions = None  # Internal private property for the lazy property self.instr_options - see the getter for self.instr_options. Initialized by the first access
		self.query_instr_status: bool = True
		self.opc_query_after_write: bool = False
		self.bin_float_numbers_format = self._settings.bin_float_numbers_format
		self.bin_int_numbers_format = self._settings.bin_int_numbers_format
		self.opc_query_after_write: bool = self._settings.opc_query_after_write
		self.stb_in_error_check: bool = self._settings.stb_in_error_check

		self.manufacturer: str = 'Rohde&Schwarz'
		self.model: str = 'R&S Instrument'
		self.serial_number: str = '100001'
		self.firmware_version: str = '1.00'

		direct_start_time = datetime.now()
		try:
			dir_str = ' from direct session' if self._direct_session else ''
			if self._simulating:
				# noinspection PyTypeChecker
				self._set_session(VisaSessionSim(resource_name, self._settings, direct_session))
				self._init_logger(self._session.resource_name)
				self._log_start_segment(direct_start_time)
				self._lock = self._session.get_lock()
				self._instr_options = InstrumentOptions.Options('K0', InstrumentOptions.ParseMode.KeepOriginal)
				self._session.write('*OPT K0')
				self._log_info('Simulation session init', f"Simulation device{dir_str} '{self.resource_name}'")
				self._log_end_segment()
				return

			self._set_session(VisaSession(resource_name, self._settings, direct_session))
			self._init_logger(self._session.resource_name)
			self._log_start_segment(direct_start_time)

			self._lock = self._session.get_lock()
			with self._lock:
				self._session.clear_before_read()
				self.idn_string = Utilities.trim_str_response(self._session.query_str(self._session.cmd_idn)).strip()
				# NRP-Z session coercing
				if self._session.is_rsnrp_session():
					self._settings.instr_options_parse_mode = InstrumentOptions.ParseMode.Skip
					self.stb_in_error_check = False
				self.instr_options_parse_mode = self._settings.instr_options_parse_mode
			self._log_info('Session init', f"Device{dir_str} '{self.resource_name}' IDN: {self.idn_string}")

		except RsInstrException as e:
			if not self.logger:
				self._assure_logger_exists()
				self._log_start_segment(direct_start_time)
			self._log_error('Session init error', e.args[0], self._start_time, datetime.now())
			raise

		finally:
			if not self.logger:
				self._assure_logger_exists()
				self._log_start_segment(direct_start_time)
			self.logger.allow_log_string_adjust = True
			self._log_end_segment()

	def __str__(self):
		if self._simulating:
			return f"Simulated, Model: '{self.model}', ResourceName: '{self.resource_name}'"
		else:
			return f"Instrument Model: '{self.model}', ResourceName: '{self.resource_name}'"

	def _init_logger(self, rsrc_name: str) -> None:
		"""Initializes the logger with default settings."""
		self.logger = ScpiLogger(rsrc_name, self._settings.encoding)
		if self._settings.logging_name is not None and self._settings.logging_name != rsrc_name:
			self.logger.device_name = self._settings.logging_name
		self.logger.allow_log_string_adjust = False
		self.logger.mode = self._settings.logging_mode
		self.logger.log_to_console = self._settings.log_to_console
		self.logger.log_to_udp = self._settings.log_to_udp
		self.logger.udp_port = self._settings.log_udp_port
		if self._settings.log_to_global_target:
			self.logger.set_logging_target_global()

	def _assure_logger_exists(self) -> None:
		"""Initializes backup logger if the VisaSession ended with exception causing main logger to not be initialized.
		If the logger instance exists, the method does nothing."""
		if self.logger:
			return
		if self._direct_session:
			session = VisaSession.get_and_check_direct_session(self._direct_session)
			self._init_logger(session.resource_name)
			return
		self._init_logger(self.resource_name)

	@property
	def _session(self) -> VisaSession:
		"""Returns instrument VISA session. Throws exception if null."""
		if self.__session is None:
			raise RsInstrException(f'Instrument session \'{self.resource_name}\' is invalid. You have probably closed it already.')
		return self.__session

	def _set_session(self, session: VisaSession) -> None:
		"""Sets the VISA session to the provided value."""
		self.__session = session

	def _clear_session(self) -> None:
		"""Sets the VISA session to None."""
		self.__session = None

	def _session_exists(self) -> bool:
		"""Returns true, if the VISA session is not None."""
		return self.__session is not None

	@property
	def allow_reconnect(self) -> bool:
		"""Returns true, if the reconnection is supported - the session is unique, and not reused."""
		return not self._direct_session

	def reconnect(self, force_close: bool = False) -> bool:
		"""If the connection is not active, the method tries to reconnect to the device
		If the connection is active, and force_close is False, the method does nothing.
		If the connection is active, and force_close is True, the method closes, and opens the session again.
		Returns True, if the reconnection has been performed."""
		if not self.allow_reconnect:
			raise RsInstrException('Reused sessions do not support reconnection')
		self._start_time = datetime.now()
		active = self.is_connection_active()
		log_info = 'Forced Reconnection' if force_close else 'Reconnection'
		if force_close and active:
			self._log_info(log_info, 'Session was active, closing the session')
			self.close(log_info)
			active = False
		if active is True:
			self._log_info(log_info, 'Session is still active, no action needed')
			return False
		if not active:
			# Connect again
			try:
				dir_str = ' from direct session' if self._direct_session else ''
				sim = ' simulation' if self._direct_session else ''
				init_method = VisaSessionSim if self._simulating else VisaSession

				self._log_start_segment()
				self.__session = init_method(self.resource_name, self._settings, self._direct_session)
				self._log_info(log_info, f"Session init,{sim} device{dir_str} '{self.resource_name}, IDN: {self.idn_string}'")
				self._log_end_segment()
				with self._lock:
					self._session.clear_before_read()
			except RsInstrException as e:
				if self.logger:
					self._log_error(log_info, e.args[0])
				raise
			finally:
				if self.logger:
					self._log_end_segment()
			return True

	def set_simulating_cmds(self) -> None:
		"""Updated cached values in the simulating VISA session to properly respond to *IDN? or *OPT?"""
		if self._simulating:
			self._session.write(f'*idn {self.idn_string}')
			self._session.write(f'*opt {Conv.list_to_csv_str(self.instr_options.get_all())}')
			self._session.write(f'*opc 1')
			self._session.write(f'*stb 0')
			self._session.write(f'*tst 0,"Passed"')
			self._session.write(f'syst:err 0,"No Error"')
			self._session.write(f'system:error 0,"No Error"')

	def get_last_sent_cmd(self) -> str:
		"""Returns the last commands sent to the instrument. Only works in simulation mode."""
		if self._simulating:
			# noinspection PyUnresolvedReferences
			return self._session.get_last_sent_cmd()
		raise RsInstrException('get_last_sent_cmd() can only be used in simulation mode')

	def assign_lock(self, lock: threading.RLock) -> None:
		"""Assigns the thread lock provided from by the user. Trickles down to the VisaSession."""
		self._lock = lock
		self._session.assign_lock(lock)

	def get_lock(self) -> threading.RLock:
		"""Returns the current RLock object."""
		return self._lock

	def clear_lock(self):
		"""Clears the existing thread lock, making the current session thread-independent from others that might share the current thread lock."""
		self.assign_lock(threading.RLock())

	def _log_start_segment(self, direct_start_time: datetime = None):
		"""Sets start time for the log entry to be able to calculate the duration. You can enter a direct start time."""
		self._last_error_log = None
		if direct_start_time:
			self._start_time = direct_start_time
		else:
			self._start_time = datetime.now()
		self.logger.start_new_segment()

	def _log_info(self, log_string_info: str, log_string: str) -> None:
		"""Logs an ASCII entry."""
		self._last_exc_log = None
		self.logger.info(self._start_time, datetime.now(), log_string_info, log_string)

	def _log_info_list(self, log_string_info: str, list_data: List) -> None:
		"""Logs a List entry."""
		self._last_exc_log = None
		self.logger.info_list(self._start_time, datetime.now(), log_string_info, list_data)

	def _log_info_bin(self, log_string_info: str, log_data: bytes) -> None:
		"""Logs a binary entry."""
		self._last_exc_log = None
		self.logger.info_bin(self._start_time, datetime.now(), log_string_info, log_data)

	def _log_info_var_stream(self, log_string_info: str, stream: StreamWriter) -> None:
		"""Logs a stream entry - must be variable only, but can be binary or ascii."""
		self._last_exc_log = None
		if stream.binary:
			self.logger.info_bin(self._start_time, datetime.now(), log_string_info, stream.content)
		else:
			self.logger.info(self._start_time, datetime.now(), log_string_info, stream.content)

	def _log_error(self, log_string_info: str, log_string: str, start_time: datetime = None, end_time: datetime = None) -> None:
		"""Logs an ASCII error entry."""
		self.logger.error(start_time, end_time, log_string_info, log_string)

	def _log_exception(self, e: Exception, context: str = None, start_time: datetime = None, end_time: datetime = None) -> None:
		"""Logs an ASCII error entry taken from the exception message."""
		if start_time is None:
			start_time = self._start_time
		if end_time is None:
			end_time = datetime.now()

		msg: str = e.args[0]
		if isinstance(e, StatusException):
			# Status exceptions are handled specially to prevent duplicate logging of the same StatusException
			if msg == self._last_exc_log:
				return
			self._last_exc_log = msg
			msg = f'{e.__class__.__name__}: {msg}'
			self.logger.error(start_time, end_time, context, msg)
		else:
			self._last_exc_log = None
			self.logger.error(start_time, end_time, context, f'{e.__class__.__name__}: {msg}')

	def _log_end_segment(self) -> None:
		"""Ends logging segment."""

		if self._start_time:
			# Accumulate the spent times
			end_time = datetime.now()
			if end_time >= self._start_time:
				diff = end_time - self._start_time
				self.total_execution_time += diff

		self._start_time = None
		self._last_exc_log = None
		self.logger.end_current_segment()

	@property
	def visa_manufacturer(self) -> str:
		"""Returns the visa manufacturer of the current session."""
		return self._session.manufacturer

	@property
	def encoding(self) -> str:
		"""Returns string<=>bytes encoding of the session."""
		return self._session.encoding

	@encoding.setter
	def encoding(self, value: str) -> None:
		"""Sets string<=>bytes encoding of the session."""
		self._session.encoding = value
		self.logger.encoding = value

	def set_link_handler(self, link_name: str, handler: Callable) -> Callable:
		"""Adds / Updates link handler for the entered link_name.
		Handler API: handler(event_args: ArgLinkedEventArgs)
		Returns the previous registered handler, or None if no handler was registered before."""
		return self._linker.set_handler(link_name, handler)

	def del_link_handler(self, link_name: str) -> Callable:
		"""Deletes link handler for the link_name.
		Returns the deleted handler, or None if none existed."""
		return self._linker.del_handler(link_name)

	def del_all_link_handlers(self) -> int:
		"""Deletes all the link handlers.
		Returns number of deleted links."""
		return self._linker.del_all_handlers()

	@property
	def idn_string(self) -> str:
		"""Returns instrument's identification string."""
		return self._idn_string

	@idn_string.setter
	def idn_string(self, value: str) -> None:
		"""IDN string. Set it to force a different IDN string than the default *IDN? response."""
		self._idn_string = value
		self._parse_idn_string(self._idn_string)

	@property
	def instr_options(self) -> InstrumentOptions:
		"""Public getter for the lazy property instr_options"""
		if self._instr_options is None:
			self._query_options_and_parse(self.instr_options_parse_mode)
		return self._instr_options

	@property
	def opc_timeout(self) -> int:
		"""See the opc_timeout.setter."""
		return self._session.opc_timeout

	@opc_timeout.setter
	def opc_timeout(self, value: int) -> None:
		"""Sets / Gets timeout in milliseconds for all the operations that use OPC synchronization."""
		self._session.opc_timeout = value

	@property
	def visa_timeout(self) -> int:
		"""See the visa_timeout.setter."""
		return self._session.visa_timeout

	@visa_timeout.setter
	def visa_timeout(self, value: int) -> None:
		"""Sets / Gets visa IO timeout in milliseconds."""
		self._session.visa_timeout = value

	@property
	def data_chunk_size(self) -> int:
		"""Returns max chunk size of one data block."""
		return self._session.data_chunk_size

	@data_chunk_size.setter
	def data_chunk_size(self, chunk_size: int) -> None:
		"""Sets the maximum size of one block transferred during write/read operations."""
		self._session.data_chunk_size = int(chunk_size)

	# noinspection PyMethodMayBeStatic
	def _sim_cached_value_found(self, value) -> bool:
		return value != 'Simulating'

	# noinspection PyMethodMayBeStatic
	def _sim_cached_value_not_found(self, value) -> bool:
		return value == 'Simulating'

	def _parse_idn_string(self, idn_string: str) -> None:
		"""Parse the *IDN? response to:
		- Manufacturer
		- Model
		- SerialNumber
		- FirmwareRevision"""
		idn_string = idn_string.strip()
		items = Utilities.trim_str_response(idn_string).rstrip(', ').split(',')
		items_count = len(items)
		if items_count < 3 or items_count > 5:
			raise RsInstrException(f"The instrument *IDN? string parsing failed. Parsed *IDN? response: '{idn_string}'")
		self.manufacturer = Utilities.trim_str_response(items[0].strip())
		self.full_model_name = Utilities.trim_str_response(items[1].strip())
		self.serial_number = Utilities.trim_str_response(items[2].strip())
		self.firmware_version = ''
		if items_count > 3:
			self.firmware_version = Utilities.trim_str_response(items[3].strip())

		self.model = self.full_model_name
		if self._settings.idn_model_full_name is False:
			m = re.search(r'([a-zA-Z ]+)([\-\da-zA-Z ]*)', self.full_model_name)
			if m:
				self.model = m.group(1)

	def fits_idn_pattern(self, patterns: List[str], supported_models: List[str]) -> None:
		"""Throws exception if the current instrument model does not fit  any of the patterns.
		The supported_models argument is only used for exception messages"""
		matches = False
		assert self._idn_string, f'*IDN? was not assigned yet.'
		for x in patterns:
			matches = re.search(x, self.idn_string, re.IGNORECASE)
			if matches:
				break
		if not matches:
			message = f"Instrument is not supported.\n*IDN? string: '{self.idn_string}'"
			if len(supported_models) > 0:
				message += f"\nSupported models: '{', '.join(supported_models)}'"
			if len(patterns) == 1:
				message += f"\nSupported IDN pattern: '{patterns[0]}'"
			if len(patterns) > 1:
				message += "\nSupported IDN patterns:\n" + '\n'.join(patterns)
			raise UnexpectedResponseException(self.resource_name, message)

	def reset_time_statistics(self) -> None:
		"""Resets all execution and total time counters.
		Changes the self.total_time_startpoint and resets the self.total_execution_time."""
		self.total_execution_time = timedelta()
		self.total_time_startpoint = datetime.now()

	def _query_options_and_parse(self, mode: InstrumentOptions.ParseMode) -> None:
		"""Queries *OPT? and parses it based on the ParseMode."""
		if mode == InstrumentOptions.ParseMode.Skip:
			self._instr_options = InstrumentOptions.Options('', mode)
			return
		if self._simulating is False:
			with self._lock:
				log_info = 'Query Instrument Options'
				try:
					self._log_start_segment()
					opts = self._session.query_str_no_tout_err('*OPT?', 1000)
					if opts is None:
						opts = 'Cannot read the instrument options - *OPT? query is not supported'
					self._instr_options = InstrumentOptions.Options(opts, mode)
					self._log_info(log_info, f'*OPT? {opts}')
				except RsInstrException as e:
					self._log_exception(e, log_info)
					raise
				finally:
					self._log_end_segment()

	def add_global_repcap(self, name: str, rep_cap: RepeatedCapability) -> None:
		"""Adds the global repcap name to the list of global repcaps
		and sets its value to the provided default value."""
		if name in self._global_repcaps:
			raise RsInstrException(f"Error adding new global repcap: '{name}' already exists in the list.")
		self._global_repcaps[name] = rep_cap

	def set_global_repcap_value(self, name: str, enum_value: Enum) -> None:
		"""Updates the existing global repcap value as enum"""
		if name not in self._global_repcaps:
			raise RsInstrException(f"Error updating global repcap: '{name}' does not exist in the list.")
		self._global_repcaps[name].set_enum_value(enum_value)

	def get_global_repcap_value(self, name: str) -> Enum:
		"""Returns the current global repcap value as enum"""
		if name not in self._global_repcaps:
			raise RsInstrException(f"Error retrieving global repcap: '{name}' does not exist in the list.")
		return self._global_repcaps[name].get_enum_value()

	def _replace_global_repcaps(self, cmd: str) -> str:
		"""Replaces all the global repcaps in the command: e.g. '<instance>' => '1'.
		Returns the replaced command."""
		for name, value in self._global_repcaps.items():
			cmd_value = value.get_cmd_string_value()
			cmd = cmd.replace(name, cmd_value)
		return cmd

	def query_opc(self, timeout: int = 0) -> bool:
		"""Sends *OPC? query and returns the result.
		If you define timeout > 0, the VISA timeout is set to that value just for this method call."""
		with self._lock:
			try:
				self._log_start_segment()
				self.start_send_read_event('*OPC?', False)
				opc: bool = self._session.query_opc(timeout)
				self.end_send_read_event()
				self._log_info('Query OPC', '1' if opc else '0')
			except RsInstrException as e:
				self._log_exception(e, 'Query OPC')
				raise
			finally:
				self._log_end_segment()
			return opc

	def query_all_syst_errors(self, include_codes: bool = True, enable_log: bool = True) -> List[str] or List[int, str] or None:
		"""Returns all errors in the instrument's error queue. If no error is detected, the return value is None.
		If include_codes is False:
			- you get List of strings with messages.
		If include_codes is True:
			- you get List of Tuples (code, message)"""
		with self._lock:
			log_info = 'Query all system errors'
			if enable_log is True:
				if self._start_time is None:
					self._start_time = datetime.now()
			try:
				self.start_send_read_event('SYST:ERROR?', False)
				errors = self._session.query_all_syst_errors()
				self.end_send_read_event()
				if errors is not None:
					if include_codes:
						entries = [f"{x[1]},'{x[0]}'" for x in errors]
					else:
						errors = [x[1] for x in errors]
						entries = errors
				# Return errors as list of strings
				if enable_log is True:
					if errors is None or len(errors) == 0:
						self._log_info(log_info, 'No errors')
					elif len(errors) == 1:
						self._log_info(log_info, f'1 error detected - {entries[0]}')
					else:
						self._log_info(log_info, f'{len(errors)} errors detected (last one on top)')
						i = 1
						for x in entries:
							self._log_info(f'SYST:ERROR? {i}', x)
							i += 1
			except RsInstrException as e:
				# General errors: log the exception message
				if enable_log is True:
					self._log_exception(e, log_info, start_time=self._start_time, end_time=datetime.now())
				raise
			return errors

	def check_status(self) -> None:
		"""Throws InstrumentStatusException in case of an error in the instrument's error queue.
		The procedure is skipped, if the QueryInstrumentStatus is set to false."""
		with self._lock:
			if not self.query_instr_status:
				return
			if self._start_time is None:
				self._start_time = datetime.now()
			try:
				call_syst_error = self._session.error_in_error_queue() if self.stb_in_error_check else True
				if call_syst_error:
					errors = self.query_all_syst_errors(enable_log=False)
					assert_no_instrument_status_errors(self.resource_name, errors)
				if self.logger.log_status_check_ok:
					self._log_info('Status check', 'OK')
			except RsInstrException as e:
				# General errors: log the exception message
				self._log_exception(e, 'Status check', start_time=self._start_time, end_time=datetime.now())
				raise

	def is_connection_active(self) -> bool:
		"""Returns true, if the VISA connection is active and the communication with the instrument still works.
		This is achieved by:
		- checking the session property timeout
		- sending the *IDN? query"""
		if self._session_exists() is False:
			return False
		return self._session.is_connection_active()

	def clear_status(self) -> None:
		"""Clears instrument's status subsystem."""
		with self._lock:
			try:
				self._log_start_segment()
				self._session.clear()
				self._session.clear_before_read()
				self._log_info('Clear status', 'OK')
			except RsInstrException as e:
				self._log_exception(e, 'Clear status')
				raise
			finally:
				self._log_end_segment()

	def reset(self, timeout: int = 0) -> None:
		"""Resets the instrument and clears its status.
		If you define timeout > 0, the VISA timeout is set to that value just for this method call."""
		old_tout = 0
		if timeout != 0:
			old_tout = self.visa_timeout
			self.visa_timeout = timeout
		try:
			with self._lock:
				self.write(self._settings.cmd_reset, True)
				self.query_opc()
				self.clear_status()
				self._log_start_segment()
				self.check_status()
				self._log_end_segment()
		finally:
			if old_tout > 0:
				self.visa_timeout = old_tout

	def write(self, cmd: str, block_callback: bool = False, log_info: str = 'Write') -> None:
		"""Writes string command to the instrument."""
		with self._lock:
			try:
				self._log_start_segment()
				cmd = self._replace_global_repcaps(cmd)
				self._call_before_write_handler(cmd, block_callback)
				self._session.write(cmd)
				if self.opc_query_after_write:
					self._session.query_opc()
				if self.on_write_handler:
					self.send_write_str_event(cmd, False)
				self._log_info(log_info, cmd)
				self.check_status()
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()

	def write_with_opc(self, cmd: str, timeout: int = None, block_callback: bool = False, log_info: str = 'Write string with OPC') -> None:
		"""Writes a OPC-synced command.
		Also performs error checking if the property self.query_instr_status is set to True.
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			try:
				self._log_start_segment()
				cmd = self._replace_global_repcaps(cmd)
				self._call_before_write_handler(cmd, block_callback)
				self._session.write_with_opc(cmd, timeout)
				self._session.query_and_clear_esr()
				if self.on_write_handler:
					self.send_write_str_event(cmd, True)
				self._log_info(log_info, cmd)
				self.check_status()
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()

	def write_struct(self, cmd: str, struct: object) -> None:
		"""Writes command to the instrument with the parameter composed of the entered structure."""
		with self._lock:
			worker = ArgStructList(struct)
			param = worker.compose_cmd_string()
			cmd += f' {param}'.rstrip()
			self.write(cmd, log_info='Write structure')

	def write_struct_with_opc(self, cmd: str, struct: object, timeout: int = None) -> None:
		"""Writes OPC-synced command to the instrument with the parameter composed of the entered structure.
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			worker = ArgStructList(struct)
			param = worker.compose_cmd_string()
			cmd += f' {param}'.rstrip()
			self.write_with_opc(cmd, timeout, log_info='Write structure with OPC')

	def query_str(self, query: str, block_callback: bool = False, log_info: str = 'Query string') -> str:
		"""Sends a query and reads response from the instrument.
		The response is trimmed of any trailing LF characters and has no length limit."""
		with self._lock:
			try:
				self._log_start_segment()
				query = self._replace_global_repcaps(query)
				self.start_send_read_event(query, False)
				self._call_pre_query_handler(query, block_callback)
				response = self._session.query_str(query)
				self.end_send_read_event()
				self._log_info(log_info, f'{query} {response}')
				self.check_status()
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()
			return response

	def query_str_with_opc(self, query: str, timeout: int = None, block_callback: bool = False, log_info: str = 'Query string with OPC') -> str:
		"""Sends a OPC-synced query.
		Also performs error checking if the self.query_instr_status is true.
		The response is trimmed of any trailing LF characters and has no length limit.
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			try:
				self._log_start_segment()
				query = self._replace_global_repcaps(query)
				self.start_send_read_event(query, True)
				self._call_pre_query_handler(query, block_callback)
				response = self._session.query_str_with_opc(query, timeout, log_info)
				self.end_send_read_event()
				self._session.query_and_clear_esr()
				self._log_info(log_info, f'{query} {response}')
				self.check_status()
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()
			return response

	def query_bin_block(self, query: str, log_info: str = 'Query binary block') -> bytes:
		"""Queries binary data block to bytes and returns data as bytes.
		Throws an exception if the returned data was not a binary data."""
		with self._lock:
			with StreamWriter.as_bin_var() as stream:
				try:
					log_info = f'{log_info} {query}'
					self._log_start_segment()
					query = self._replace_global_repcaps(query)
					self.start_send_read_event(query, False)
					self._call_pre_query_handler(query, False)
					self._session.query_bin_block(query, stream, True)
					self.end_send_read_event()
					self._log_info_var_stream(f'{log_info}, received', stream)
					self.check_status()
				except RsInstrException as e:
					self._log_exception(e, log_info)
					raise
				finally:
					self._log_end_segment()
				return stream.content

	def query_bin_block_with_opc(self, query: str, timeout: int = None, log_info: str = 'Query binary block with OPC') -> bytes:
		"""Sends a OPC-synced query and returns data as bytes.
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			with StreamWriter.as_bin_var() as stream:
				try:
					log_info = f'{log_info} {query}'
					self._log_start_segment()
					query = self._replace_global_repcaps(query)
					self.start_send_read_event(query, True)
					self._call_pre_query_handler(query, False)
					self._session.query_bin_block_with_opc(query, stream, True, timeout)
					self.end_send_read_event()
					self._session.query_and_clear_esr()
					self._log_info_var_stream(f'{log_info}, received', stream)
					self.check_status()
				except RsInstrException as e:
					self._log_exception(e, log_info)
					raise
				finally:
					self._log_end_segment()
				return stream.content

	def query_bin_block_to_file(self, query: str, file_path: str, append: bool = False, log_info='Query binary block to file') -> None:
		"""Queries binary data block to the provided file.
		If append is False, any existing file content is discarded.
		If append is True, the new content is added to the end of the existing file, or if the file does not exit, it is created.
		Throws an exception if the returned data was not a binary data."""
		with self._lock:
			with StreamWriter.as_bin_file(file_path, append) as stream:
				try:
					self._log_start_segment()
					query = self._replace_global_repcaps(query)
					self.start_send_read_event(query, False)
					self._call_pre_query_handler(query, False)
					self._session.query_bin_block(query, stream, True)
					self.end_send_read_event()
					add_str = 'target file' if append is False else 'appended to target file'
					self._log_info(log_info, f'Query {query} - written {size_to_kb_mb_string(stream.written_len, True)}, {add_str} {file_path}')
					self.check_status()
				except RsInstrException as e:
					self._log_exception(e, log_info)
					raise
				finally:
					self._log_end_segment()

	def query_bin_block_to_file_with_opc(self, query: str, file_path: str, append: bool = False, timeout: int = None, log_info='Query binary block to file with OPC') -> None:
		"""Sends a OPC-synced query and writes the returned data to the provided file.
		If append is False, any existing file content is discarded.
		If append is True, the new content is added to the end of the existing file, or if the file does not exit, it is created.
		Throws an exception if the returned data was not a binary data."""
		with self._lock:
			with StreamWriter.as_bin_file(file_path, append) as stream:
				try:

					self._log_start_segment()
					query = self._replace_global_repcaps(query)
					self.start_send_read_event(query, True)
					self._call_pre_query_handler(query, False)
					self._session.query_bin_block_with_opc(query, stream, True, timeout)
					self.end_send_read_event()
					self._session.query_and_clear_esr()
					add_str = 'target file' if append is False else 'appended to target file'
					self._log_info(log_info, f'Query {query} - written {size_to_kb_mb_string(stream.written_len, True)}, {add_str} {file_path}')
					self.check_status()
				except RsInstrException as e:
					self._log_exception(e, log_info)
					raise
				finally:
					self._log_end_segment()

	def query_int(self, query: str) -> int:
		"""Sends a query and reads response from the instrument as integer."""
		with self._lock:
			string = self.query_str(query, log_info='Query integer')
			if self._simulating and self._sim_cached_value_not_found(string):
				return 0
			return Conv.str_to_int(string)

	def query_int_with_opc(self, query: str, timeout: int = None) -> int:
		"""Sends a OPC-synced query and reads response from the instrument as integer number.
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			string = self.query_str_with_opc(query, timeout, log_info='Query integer with OPC')
			if self._simulating and self._sim_cached_value_not_found(string):
				return 0
			return Conv.str_to_int(string)

	def query_float(self, query: str) -> float:
		"""Sends a query and reads response from the instrument as float number."""
		with self._lock:
			string = self.query_str(query, log_info='Query float')
			if self._simulating and self._sim_cached_value_not_found(string):
				return 0.0
			return Conv.str_to_float(string)

	def query_float_with_opc(self, query: str, timeout: int = None) -> float:
		"""Sends a OPC-synced query and reads response from the instrument as float number.
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			string = self.query_str_with_opc(query, timeout, log_info='Query float with OPC')
			if self._simulating and self._sim_cached_value_not_found(string):
				return 0.0
			return Conv.str_to_float(string)

	def query_bool(self, query: str) -> bool:
		"""Sends a query and reads response from the instrument as boolean value."""
		with self._lock:
			string = self.query_str(query, log_info='Query boolean')
			if self._simulating and self._sim_cached_value_not_found(string):
				return False
			return Conv.str_to_bool(string)

	def query_bool_with_opc(self, query: str, timeout: int = None) -> bool:
		"""Sends a OPC-synced query and reads response from the instrument as boolean value.
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			string = self.query_str_with_opc(query, timeout, log_info='Query boolean with OPC')
			if self._simulating and self._sim_cached_value_not_found(string):
				return False
			return Conv.str_to_bool(string)

	def query_str_list(self, query: str) -> List[str]:
		"""Sends a query and reads response from the instrument as csv-list."""
		with self._lock:
			string = self.query_str(query, log_info='Query string list')
			if self._simulating and self._sim_cached_value_not_found(string):
				string = 'AAA,BBB,CCC,DDD,EEE,FFF,GGG,HHH,III,JJJ'
			response = [trim_str_response(x) for x in string.split(',')]
			return response

	def query_str_list_with_opc(self, query: str, timeout: int = None) -> List[str]:
		"""Sends a OPC-synced query and reads response from the instrument as csv-list.
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			string = self.query_str_with_opc(query, timeout, log_info='Query string list with OPC')
			if self._simulating and self._sim_cached_value_not_found(string):
				string = 'AAA,BBB,CCC,DDD,EEE,FFF,GGG,HHH,III,JJJ'
			response = [trim_str_response(x) for x in string.split(',')]
			return response

	def query_bool_list(self, query: str) -> List[bool]:
		"""Sends a query and reads response from the instrument as csv-list of booleans."""
		with self._lock:
			string = self.query_str(query, log_info='Query boolean list')
			if self._simulating and self._sim_cached_value_not_found(string):
				string = 'True,False,0,1,1,True,true,false,true,false'
			response = [Conv.str_to_bool(x) for x in string.split(',')]
			return response

	def query_bool_list_with_opc(self, query: str, timeout: int = None) -> List[bool]:
		"""Sends a OPC-synced query and reads response from the instrument as csv-list of booleans.
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			string = self.query_str_with_opc(query, timeout, log_info='Query boolean list with OPC')
			if self._simulating and self._sim_cached_value_not_found(string):
				string = 'True,False,0,1,1,True,true,false,true,false'
			response = [Conv.str_to_bool(x) for x in string.split(',')]
			return response

	def write_bin_block(self, cmd: str, payload: bytes, log_info: str = 'Write binary block') -> None:
		"""Writes all the payload as binary data block to the instrument.
		The binary data header is added at the beginning of the transmission automatically, do not include it in the payload!!!"""
		with self._lock:
			try:
				self._log_start_segment()
				cmd = self._replace_global_repcaps(cmd)
				self._call_before_write_handler(cmd, False)
				stream = StreamReader.as_bin_var(payload)
				if self.on_write_handler:
					self.start_send_write_bin_event(cmd)
				self._session.write_bin_block(cmd, stream)
				self.end_send_write_bin_event()
				self._log_info_bin(f'{log_info} {cmd}, binary data', payload)
				self.check_status()
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()

	def write_bin_block_from_file(self, cmd: str, file_path: str, log_info: str = 'Write binary block from file') -> None:
		"""Writes all the file content as binary data block to the instrument.
		The binary data header is added at the beginning of the transmission automatically, do not include it in the file content!!!"""
		with self._lock:
			with StreamReader.as_bin_file(file_path) as stream:
				try:
					self._log_start_segment()
					cmd = self._replace_global_repcaps(cmd)
					self._call_before_write_handler(cmd, False)
					if self.on_write_handler:
						self.start_send_write_bin_event(cmd)
					self._session.write_bin_block(cmd, stream)
					self.end_send_write_bin_event()
					self._log_info(log_info, f'Command {cmd} - written {size_to_kb_mb_string(stream.read_len, True)}, source file {file_path}')
					self.check_status()
				except RsInstrException as e:
					self._log_exception(e, log_info)
					raise
				finally:
					self._log_end_segment()

	def send_file_from_pc_to_instrument(self, source_pc_file: str, target_instr_file: str) -> None:
		"""SCPI Command: MMEM:DATA \n
		Sends file from PC to the instrument"""
		cmd = f"MMEM:DATA '{target_instr_file}',"
		self.write_bin_block_from_file(cmd, source_pc_file, log_info='Send file from PC to the instrument')

	def read_file_from_instrument_to_pc(self, source_instr_file: str, target_pc_file: str, append_to_pc_file: bool = False) -> None:
		"""SCPI Command: MMEM:DATA? \n
		Reads file from instrument to the PC. \n
		Set the append_to_pc_file to True if you want to append the read content to the end of the existing PC file"""
		cmd = f"MMEM:DATA? '{source_instr_file}'"
		self.query_bin_block_to_file(cmd, target_pc_file, append_to_pc_file, log_info='Read file from instrument to the PC')

	def query_bin_or_ascii_float_list(self, query: str, log_info: str = 'Query binary or ascii float list') -> List[float]:
		"""Queries a list of floating-point numbers that can be read in ASCII format or in binary format.
		- For ASCII format, the list numbers are decoded as comma-separated values.
		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32)."""
		with self._lock:
			try:
				log_info = f'{log_info} {query}'
				self._log_start_segment()
				query = self._replace_global_repcaps(query)
				self.start_send_read_event(query, False)
				stream = StreamWriter.as_bin_var()
				self._call_pre_query_handler(query, False)
				self._session.query_bin_block(query, stream, False)
				self.end_send_read_event()
				if self._simulating and not self._session.cached_to_stream:
					return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
				if stream.binary:
					result = Conv.bytes_to_list_of_floats(stream.content, self.bin_float_numbers_format)
					self._log_info_list(f'{log_info}, received binary format list {size_to_kb_mb_string(stream.written_len, True)} {stream.written_len // len(result)} bytes per number', result)
				else:
					result = Conv.str_to_float_list(stream.content)
					self._log_info_list(f'{log_info}, received ascii format list', result)
				self.check_status()
				return result
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()

	def query_bin_or_ascii_float_list_with_opc(self, query: str, timeout: int = None, log_info: str = 'Query binary or ascii float list with OPC') -> List[float]:
		"""Sends a OPC-synced query and reads a list of floating-point numbers that can be read in ASCII format or in binary format.
		- For ASCII format, the list numbers are decoded as comma-separated values.
		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32).
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			try:
				log_info = f'{log_info} {query}'
				self._log_start_segment()
				query = self._replace_global_repcaps(query)
				self.start_send_read_event(query, True)
				stream = StreamWriter.as_bin_var()
				self._call_pre_query_handler(query, False)
				self._session.query_bin_block_with_opc(query, stream, False, timeout)
				self.end_send_read_event()
				if self._simulating and not self._session.cached_to_stream:
					return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
				if stream.binary:
					result = Conv.bytes_to_list_of_floats(stream.content, self.bin_float_numbers_format)
					self._log_info_list(f'{log_info}, received binary format list {size_to_kb_mb_string(stream.written_len, True)} {stream.written_len // len(result)} bytes per number', result)
				else:
					result = Conv.str_to_float_list(stream.content)
					self._log_info_list(f'{log_info}, received ascii format list', result)
				self._session.query_and_clear_esr()
				self.check_status()
				return result
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()

	def query_bin_or_ascii_float_list_suppressed(self, query: str, suppressed: ArgSingleSuppressed) -> List[float]:
		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument as list of floats.
		The current implementation allows for the rest of the string to be only ASCII format."""
		with self._lock:
			string = self.query_str(query, log_info='Query float list suppressed')
			if self._simulating and self._sim_cached_value_not_found(string):
				return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
			response = self._linker.cut_from_response_string(suppressed, string, query)
			return Conv.str_to_float_list(response)

	def query_bin_or_ascii_float_list_suppressed_with_opc(self, query: str, suppressed: ArgSingleSuppressed, timeout: int = None) -> List[float]:
		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument as list of floats.
		If you do not provide timeout, the method uses current opc_timeout.
		The current implementation allows for the rest of the string to be only ASCII format."""
		with self._lock:
			string = self.query_str_with_opc(query, timeout, log_info='Query float list suppressed with OPC')
			if self._simulating and self._sim_cached_value_not_found(string):
				return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
			response = self._linker.cut_from_response_string(suppressed, string, query)
			return Conv.str_to_float_list(response)

	def query_bin_or_ascii_int_list(self, query: str, log_info: str = 'Query binary or ascii integer list') -> List[int]:
		"""Queries a list of integer numbers that can be read in ASCII format or in binary format.
		- For ASCII format, the list numbers are decoded as comma-separated values.
		- For Binary Format, the numbers are decoded based on the property BinIntFormat, usually int 32-bit (FORM REAL,32)."""
		with self._lock:
			try:
				log_info = f'{log_info} {query}'
				self._log_start_segment()
				query = self._replace_global_repcaps(query)
				self.start_send_read_event(query, False)
				stream = StreamWriter.as_bin_var()
				self._call_pre_query_handler(query, False)
				self._session.query_bin_block(query, stream, False)
				self.end_send_read_event()
				if self._simulating and not self._session.cached_to_stream:
					return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
				if stream.binary:
					result = Conv.bytes_to_list_of_integers(stream.content, self.bin_int_numbers_format)
					self._log_info_list(f'{log_info}, received binary format list {size_to_kb_mb_string(stream.written_len, True)} {stream.written_len // len(result)} bytes per number', result)
				else:
					result = Conv.str_to_int_list(stream.content)
					self._log_info_list(f'{log_info}, received ascii format list', result)
				self.check_status()
				return result
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()

	def query_bin_or_ascii_int_list_with_opc(self, query: str, timeout: int = None, log_info: str = 'Query binary or ascii integer list with OPC') -> List[int]:
		"""Sends a OPC-synced query and reads a list of integer numbers that can be read in ASCII format or in binary format.
		- For ASCII format, the list numbers are decoded as comma-separated values.
		- For Binary Format, the numbers are decoded based on the property BinIntFormat, usually int 32-bit (FORM REAL,32).
		If you do not provide timeout, the method uses current opc_timeout."""
		with self._lock:
			try:
				log_info = f'{log_info} {query}'
				self._log_start_segment()
				query = self._replace_global_repcaps(query)
				self.start_send_read_event(query, True)
				stream = StreamWriter.as_bin_var()
				self._call_pre_query_handler(query, False)
				self._session.query_bin_block_with_opc(query, stream, False, timeout)
				self.end_send_read_event()
				if self._simulating and not self._session.cached_to_stream:
					return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
				if stream.binary:
					result = Conv.bytes_to_list_of_integers(stream.content, self.bin_int_numbers_format)
					self._log_info_list(f'{log_info}, received binary format list {size_to_kb_mb_string(stream.written_len, True)} {stream.written_len // len(result)} bytes per number', result)
				else:
					result = Conv.str_to_int_list(stream.content)
					self._log_info_list(f'{log_info}, received ascii format list', result)
				self._session.query_and_clear_esr()
				self.check_status()
				return result
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()

	def query_bin_or_ascii_int_list_suppressed(self, query: str, suppressed: ArgSingleSuppressed) -> List[int]:
		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument as list of integers.
		The current implementation allows for the rest of the string to be only ASCII format."""
		with self._lock:
			response = self.query_str(query, log_info='Query integer list suppressed')
			if self._simulating:
				return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
			response = self._linker.cut_from_response_string(suppressed, response, query)
			return Conv.str_to_int_list(response)

	def query_bin_or_ascii_int_list_suppressed_with_opc(self, query: str, suppressed: ArgSingleSuppressed, timeout: int = None) -> List[int]:
		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument as list of integers.
		If you do not provide timeout, the method uses current opc_timeout.
		The current implementation allows for the rest of the string to be only ASCII format."""
		with self._lock:
			response = self.query_str_with_opc(query, timeout, log_info='Query integer list suppressed with OPC')
			if self._simulating:
				return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
			response = self._linker.cut_from_response_string(suppressed, response, query)
			return Conv.str_to_int_list(response)

	def query_struct(self, query: str, struct: object) -> object:
		"""Queries string of from instrument, and parses it based on the provided structure object.
		THe method returns the copy of the entered object that it had modified."""
		with self._lock:
			string = self.query_str(query, log_info='Query structure')
			if self._simulating and self._sim_cached_value_not_found(string):
				return struct
			struct_list = ArgStructList(struct)
			struct_list.parse_from_cmd_response(string)
			self._linker.invoke_struct_intern_links(struct, struct_list.args, query)
			return struct

	def query_struct_with_opc(self, query: str, struct: object, timeout: int = None) -> object:
		"""Queries string of from instrument, and parses it based on the provided structure object.
		THe method returns the copy of the entered object that it had modified."""
		with self._lock:
			string = self.query_str_with_opc(query, timeout, log_info='Query structure with OPC')
			if self._simulating and self._sim_cached_value_not_found(string):
				return struct
			struct_list = ArgStructList(struct)
			struct_list.parse_from_cmd_response(string)
			self._linker.invoke_struct_intern_links(struct, struct_list.args, query)
			return struct

	def query_str_suppressed(self, query: str, suppressed: ArgSingleSuppressed) -> str:
		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument."""
		with self._lock:
			string = self.query_str(query, log_info='Query string suppressed')
			if self._simulating and self._sim_cached_value_not_found(string):
				return string
			response = self._linker.cut_from_response_string(suppressed, string, query)
			return response

	def query_str_suppressed_with_opc(self, query: str, suppressed: ArgSingleSuppressed, timeout: int = None) -> str:
		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument."""
		with self._lock:
			string = self.query_str_with_opc(query, timeout, log_info='Query string suppressed with OPC')
			if self._simulating and self._sim_cached_value_not_found(string):
				return string
			response = self._linker.cut_from_response_string(suppressed, string, query)
			return response

	def self_test(self, timeout: int = None) -> Tuple[int, str]:
		"""Performs instrument's selftest (*TST?).
		Returns tuple (code:int, message: str). . Code 0 means the self-test passed.
		You can define the custom timeout in milliseconds. If you do not define it, the default selftest timeout is used (usually 60 secs)."""
		with self._lock:
			if timeout is None or timeout == 0:
				timeout = self._settings.selftest_timeout
			response = self.query_str_with_opc('*TST?', timeout, log_info='Self Test')
			m = re.search(r'^(-?\d+)(,(.*))?', response)
			if not m:
				raise UnexpectedResponseException(self.resource_name, f"Unexpected response to a '*TST?' self-test query: '{response}'")
			code = Conv.str_to_int(m.group(1))
			msg = Utilities.trim_str_response(m.group(3))
			self.check_status()
			return code, msg

	def get_session_handle(self) -> object:
		"""Returns the underlying pyvisa session."""
		return self._session.get_session_handle()

	def close(self, log_info='Close') -> None:
		"""Closes the Instrument session."""
		with self._lock:
			try:
				self._log_start_segment()
				if self._session_exists():
					reused = self._session.reusing_session
					self._session.close()
					self._clear_session()
					try:
						# Tolerate error when trying to log to the closed stream.
						# This is only tolerated in the close() method.
						log_string = 'Closing reused session' if reused else 'Closing session'
						self._log_info(log_info, log_string)
					except RsInstrException as e:
						if 'Error logging to the stream' not in e.args[0]:
							raise
			except RsInstrException as e:
				self._log_exception(e, log_info)
				raise
			finally:
				self._log_end_segment()

	# Events part -------------------------------------------------------------
	@property
	def io_events_include_data(self) -> bool:
		"""If true, the read and write handlers also include read and written data."""
		return self._io_events_include_data

	@io_events_include_data.setter
	def io_events_include_data(self, value: bool) -> None:
		"""If true, the read and write handlers also include read and written data."""
		self._io_events_include_data = value
		self._session.io_events_include_data = value

	def event_args_append_instr_info(self, args: IoTransferEventArgs) -> IoTransferEventArgs:
		"""Appends instrument-related information to the read/write event argument"""
		args.chunk_size = self.data_chunk_size
		args.resource_name = self.resource_name
		return args

	def send_write_str_event(self, cmd: str, opc_sync: bool) -> None:
		"""Creates and sends write string event. The transfer is marked as done (end_of_transfer = True)."""
		args = IoTransferEventArgs.write_str(opc_sync, len(cmd), cmd)
		args.transferred_size = args.total_size
		args.chunk_ix = 0
		args.end_of_transfer = True
		args = self.event_args_append_instr_info(args)
		if self._io_events_include_data:
			args.data = cmd
		self.on_write_handler(args)

	def start_send_read_event(self, query: str, opc_sync: bool) -> None:
		"""Registers VisaSession.on_read_chunk_handler() which then generates events with each chunk transfer.
		Event handler for these events is the local function of this method, which sends IoTransferEventArgs further up to the Instrument.on_read_handler()"""
		if not self.on_read_handler:
			return

		def _read_chunk_handler(visa_args: EventArgsChunk) -> None:
			"""Receives events from VisaSession on read chunk transfers, and sends them as IoTransferEventArgs to the Instrument.on_read_handler()"""
			args.end_of_transfer = visa_args.end_of_transfer
			args.chunk_ix = visa_args.chunk_ix
			args.total_chunks = visa_args.total_chunks
			args.chunk_size = visa_args.chunk_size
			args.transferred_size = visa_args.transferred_size
			args.total_size = visa_args.total_size
			args.data = visa_args.data
			args.binary = visa_args.binary
			self.on_read_handler(args)

		args = IoTransferEventArgs.read_chunk(opc_sync, query)
		args = self.event_args_append_instr_info(args)
		self._session.on_read_chunk_handler = _read_chunk_handler

	def end_send_read_event(self):
		"""Unregisters VisaSession.on_read_chunk_handler()"""
		self._session.on_read_chunk_handler = None

	def start_send_write_bin_event(self, cmd: str) -> None:
		"""Registers VisaSession.on_write_chunk_handler() which then generates events with each chunk transfer.
		Event handler for these events is the local function of this method, which sends IoTransferEventArgs further up to the Instrument.on_write_handler()"""
		if not self.on_write_handler:
			return

		def _write_chunk_handler(visa_args: EventArgsChunk) -> None:
			"""Receives events from VisaSession on write chunk transfers, and sends them as IoTransferEventArgs to the Instrument.on_write_handler()"""
			args.end_of_transfer = visa_args.end_of_transfer
			args.chunk_ix = visa_args.chunk_ix
			args.total_chunks = visa_args.total_chunks
			args.chunk_size = visa_args.chunk_size
			args.transferred_size = visa_args.transferred_size
			args.total_size = visa_args.total_size
			args.data = visa_args.data
			args.binary = visa_args.binary
			self.on_write_handler(args)

		args = IoTransferEventArgs.write_bin(cmd)
		args = self.event_args_append_instr_info(args)
		self._session.on_write_chunk_handler = _write_chunk_handler

	def end_send_write_bin_event(self):
		"""Unregisters VisaSession.on_write_chunk_handler()"""
		self._session.on_write_chunk_handler = None

	def _call_before_write_handler(self, cmd: str, block_callback: bool) -> None:
		"""Calls the _pre_write_handler if defined. Used in all the base write methods."""
		if block_callback is False and self._before_write_handler:
			self._before_write_handler(self, cmd)

	def _call_pre_query_handler(self, query: str, block_callback: bool) -> None:
		"""Calls the _pre_query_handler if defined. Used in all the base query methods."""
		if block_callback is False and self._before_query_handler:
			self._before_query_handler(self, query)

	@property
	def before_write_handler(self) -> Callable:
		"""Returns the handler of before_write events. \n
		:return: current before_write_handler"""
		return self._before_write_handler

	@before_write_handler.setter
	def before_write_handler(self, handler: Callable) -> None:
		"""Sets handler for before_write events.
		The before_write event is invoked before each write operation (only once, not for every chunk)
		Event prototype: handler(io: Instrument, cmd: str)
		:param handler: new handler"""
		self._before_write_handler = handler

	@property
	def before_query_handler(self) -> Callable:
		"""Returns the handler of before_query events. \n
		:return: current before_query_handler"""
		return self._before_query_handler

	@before_query_handler.setter
	def before_query_handler(self, handler: Callable) -> None:
		"""Sets handler for before_query events.
		The before_query event is invoked before each query operation (only once, not for every chunk)
		Event prototype: handler(io: Instrument, query: str)
		:param handler: new handler"""
		self._before_query_handler = handler



"""Data type class for the variables containing all the methods related to data types."""

from enum import Enum, auto
from typing import Any


# noinspection PyArgumentList
class DataType(Enum):
	"""Data type of variable in the driver."""
	String = auto()
	RawString = auto()
	Integer = auto()
	IntegerExt = auto()
	Boolean = auto()
	Float = auto()
	FloatExt = auto()
	Enum = auto()
	EnumExt = auto()
	StringList = auto()
	RawStringList = auto()
	IntegerList = auto()
	IntegerExtList = auto()
	BooleanList = auto()
	FloatList = auto()
	FloatExtList = auto()
	EnumList = auto()
	EnumExtList = auto()

	@property
	def is_list(self) -> bool:
		"""Returns True, if the data type is a list."""
		return self in frozenset(
			{
				DataType.StringList,
				DataType.RawStringList,
				DataType.IntegerList,
				DataType.IntegerExtList,
				DataType.BooleanList,
				DataType.FloatList,
				DataType.FloatExtList,
				DataType.EnumList,
				DataType.EnumExtList
			})

	@property
	def is_scalar(self) -> bool:
		"""Returns True, if the data type is a scalar."""
		return not self.is_list

	@property
	def is_scalar_enum(self) -> bool:
		"""Returns True, if the data type is a scalar enum or enum_ext."""
		return self == DataType.Enum or self == DataType.EnumExt

	@property
	def is_list_enum(self) -> bool:
		"""Returns True, if the data type is a list enum or list enum_ext."""
		return self == DataType.EnumList or self == DataType.EnumExtList

	@property
	def is_enum(self) -> bool:
		"""Returns True, if the data type is enum or enum array - including the extended."""
		return self in [DataType.Enum, DataType.EnumExt, DataType.EnumList, DataType.EnumExtList]

	@property
	def is_raw_string(self) -> bool:
		"""Returns True for raw string and raw string list."""
		return self == DataType.RawString or self == DataType.RawStringList

	@property
	def is_boolean(self) -> bool:
		"""Returns True for boolean and boolean list."""
		return self == DataType.Boolean or self == DataType.BooleanList

	@property
	def is_string(self) -> bool:
		"""Returns True for string and string list."""
		return self == DataType.String or self == DataType.StringList

	@property
	def element_type(self):
		"""For lists, the property returns type of the element.
		For scalars, it returns the same type."""
		if self.is_scalar:
			return self
		elif self == DataType.StringList:
			return DataType.String
		elif self == DataType.RawStringList:
			return DataType.RawString
		elif self == DataType.RawStringList:
			return DataType.RawString
		elif self == DataType.BooleanList:
			return DataType.Boolean
		elif self == DataType.IntegerList:
			return DataType.Integer
		elif self == DataType.IntegerExtList:
			return DataType.IntegerExt
		elif self == DataType.FloatList:
			return DataType.Float
		elif self == DataType.FloatExtList:
			return DataType.FloatExt
		elif self == DataType.EnumList:
			return DataType.Enum
		elif self == DataType.EnumExtList:
			return DataType.EnumExt

	def get_default_value(self, enm: Enum = None) -> Any:
		"""Returns default value for the current type.
		If the data type is Enum or EnumString, you have to provide the enum class."""
		if self.is_list:
			return []
		if self == DataType.RawString:
			return ''
		elif self == DataType.String:
			return ''
		elif self == DataType.Boolean:
			return False
		elif self == DataType.Integer:
			return 0
		elif self == DataType.IntegerExt:
			return 0
		elif self == DataType.Float:
			return 0.0
		elif self == DataType.FloatExt:
			return 0.0
		elif self == DataType.Enum:
			return enm(0)
		elif self == DataType.EnumExt:
			return enm(0)



"""Global package data that are valid for all of its instances.
Used to access the RsInstrument class variables."""

from datetime import datetime


class GlobalData:
    """Global package data that are valid for all of its instances."""

    bounded_class = None

    @classmethod
    def is_bounded(cls) -> bool:
        """Returns true, if the class is bounded to a global source."""
        return cls.bounded_class is not None

    @classmethod
    def set_logging_target(cls, value: datetime or None) -> None:
        """Sets the class variable to the entered value."""
        setattr(cls.bounded_class, '_global_logging_target_stream', value)

    @classmethod
    def get_logging_target(cls):
        """Returns the class variable value."""
        if not cls.is_bounded():
            return None
        return getattr(cls.bounded_class, '_global_logging_target_stream')

    @classmethod
    def set_logging_relative_timestamp(cls, value: datetime or None) -> None:
        """Sets the class variable to the entered value."""
        setattr(cls.bounded_class, '_global_logging_relative_timestamp', value)

    @classmethod
    def get_logging_relative_timestamp(cls) -> datetime or None:
        """Returns the class variable value."""
        if not cls.is_bounded():
            return None
        return getattr(cls.bounded_class, '_global_logging_relative_timestamp')



"""See the docstring for the RepeatedCapability class."""

from enum import Enum


# Command integer value that signals Default value "DEFAULT"
VALUE_DEFAULT = -1

# Command integer value that signals "EMPTY"
VALUE_EMPTY = -2


class RepeatedCapability(object):
	"""Represents Repeated Capability value and type"""

	def __init__(self, header_name: str, method_get_name: str, method_set_name: str, start_value: Enum):
		"""Constructor with header name, group property name and the start value"""

		self._enum_value = None
		self.enum_type = start_value.__class__
		self.name = self.enum_type.__name__
		self.header_name = header_name
		self.method_get_name = method_get_name
		self.method_set_name = method_set_name
		self._start_value = start_value
		self.set_to_start_value()

	def __str__(self) -> str:
		out = f'RepCap {self.name}'
		if self._enum_value is not None:
			out += f" = {self._enum_value}"
		return out

	@classmethod
	def clsm_assert_type(cls, enum_value: Enum or int, enum_type) -> None:
		"""Static assertion function to check if the entered value is a member of the defined repcap enum.
		In addition, the integer value is also supported."""
		if isinstance(enum_value, int):
			return
		if not isinstance(enum_value, enum_type):
			raise TypeError(f"RepCap value must be of type '{enum_type}'. Entered value type: {type(enum_value)}, value '{enum_value}'")

	@classmethod
	def clsm_get_direct_cmd_value_int(cls, enum_value: Enum or int, enum_type) -> int:
		"""Static function to get an integer interpretation of a direct enum value
		Does not work with Empty or Default"""
		RepeatedCapability.clsm_assert_type(enum_value, enum_type)
		if isinstance(enum_value, int):
			return enum_value
		return enum_value.value

	@classmethod
	def clsm_is_default_value(cls, enum_value: Enum or int, enum_type) -> bool:
		"""Returns True, if the entered value is enum.Default"""
		return cls.clsm_get_direct_cmd_value_int(enum_value, enum_type) == VALUE_DEFAULT

	def is_default_value(self) -> bool:
		"""Returns True, if the repcap value is enum.Default"""
		return RepeatedCapability.clsm_is_default_value(self._enum_value, self.enum_type)

	def set_enum_value(self, enum_value: Enum or int) -> None:
		"""Sets new enum value. Can not be Default"""
		if RepeatedCapability.clsm_is_default_value(enum_value, self.enum_type):
			raise ValueError(f"Setting RepCap enum value '{enum_value}' is not allowed. Please select a concrete value")
		if isinstance(enum_value, int):
			# Find the enum value that corresponds to the entered integer value
			enum_value = self.enum_type(enum_value)
		self._enum_value = enum_value

	def get_enum_value(self) -> Enum:
		"""Returns the actual enum value"""
		return self._enum_value

	def set_to_start_value(self) -> None:
		"""Sets back to the value entered in the constructor"""
		self.set_enum_value(self._start_value)

	def matches_type(self, enum_type) -> bool:
		"""Returns true, if the entered type matches the EnumType"""
		return self.enum_type == enum_type

	@classmethod
	def clsm_get_cmd_string_value(cls, enum_value: Enum or int, enum_type) -> str:
		"""Class method version of the get_cmd_string_value().
		Converts RepCap integer value to string
		ValueEmpty is converted to "" (Not valid, but tolerated)
		ValueDefault throws an exception
		0 is converted to "" (Not valid, but tolerated)
		Positive numbers are converted to integer strings e.g. 1 => '1' """
		number = cls.clsm_get_direct_cmd_value_int(enum_value, enum_type)
		if number == VALUE_EMPTY:
			return ''
		if number == 0:
			# return empty string, if the enum definition does not contain a valid element with value 0
			enum_values = [x.value for x in enum_type]
			return '0' if 0 in enum_values else ''
		if number == VALUE_DEFAULT:
			raise ValueError(f"RepCap enum value Default can not be converted to the command string value. RepCap: {enum_type}")
		return str(number)

	def get_cmd_string_value(self) -> str:
		"""Converts RepCap integer value to string
		ValueEmpty is converted to "" (Not valid, but tolerated)
		ValueDefault throws an exception
		0 is converted to "" (Not valid, but tolerated)
		Positive numbers are converted to integer strings e.g. 1 => '1' """
		return RepeatedCapability.clsm_get_cmd_string_value(self._enum_value, self.enum_type)



"""See the docstring for the StreamReader class."""

from enum import Enum
from os import path
from typing import AnyStr

from .Utilities import size_to_kb_mb_string
from .InstrumentErrors import RsInstrException


class Type(Enum):
	"""Defines type of the stream - variable or file."""
	Variable = 1
	File = 2


class StreamReader:
	"""Lightweight stream reader implementation. Data source can be: \n
	- variable
	- file"""

	def __init__(self, binary: bool, source: Type, data: AnyStr):
		"""Initializes StreamReader instance.\n
		:param binary: True: Binary data, False: ASCII data
		:param source: Source type for the stream. Variable / File
		:param data: Depending on the 'binary' and 'source':
		For source type Variable the data must be either bytes() or str
		For source type File data must be string with existing file path."""
		self._source = source
		self._binary = binary
		self._start_ptr = 0
		self._read_len = 0

		if self._source == Type.Variable:
			if self._binary:
				assert isinstance(data, bytes), f'Data must be of bytes type. Actual type: {type(data)}'
			else:
				assert isinstance(data, str), f'Data must be of string type. Actual type: {type(data)}'
			self._data = data
			self._full_len = len(self._data)
		elif self._source == Type.File:
			assert isinstance(data, str), f'Data must be of string type (file path). Actual type: {type(data)}'
			if not path.isfile(data):
				raise RsInstrException(f'File does not exist. File path: {data}')
			self.file_path = data
			self._data = open(self.file_path, 'rb' if self._binary else 'r')
			self._full_len = path.getsize(self.file_path)
		else:
			raise RsInstrException(f'StreamReader unknown type {source}')

	@classmethod
	def as_bin_var(cls, data: bytes) -> 'StreamReader':
		"""Creates new StreamReader from bytes.
		:param data: [bytes] data for the stream."""
		return cls(True, Type.Variable, data)

	@classmethod
	def as_string_var(cls, data: str) -> 'StreamReader':
		"""Creates new StreamReader from string.
		:param data: [str] data for the stream."""
		return cls(False, Type.Variable, data)

	@classmethod
	def as_bin_file(cls, file_path: str) -> 'StreamReader':
		"""Creates new StreamReader from binary file. The file must exist at this time.
		:param file_path: [str] Path to the file."""
		return cls(True, Type.File, file_path)

	@classmethod
	def as_text_file(cls, file_path: str) -> 'StreamReader':
		"""Creates new StreamReader from text file. The file must exist at this time.
		:param file_path: [str] Path to the file."""
		return cls(False, Type.File, file_path)

	def __str__(self):
		if self._source == Type.Variable:
			mode = 'binary' if self._binary else 'string'
			return f'StreamReader {mode} data, full size {size_to_kb_mb_string(self._full_len, True)}, remaining size {size_to_kb_mb_string(len(self), True)}'
		if self._source == Type.File:
			mode = 'binary' if self._binary else 'text'
			return f'StreamReader {mode}, full size {size_to_kb_mb_string(self._full_len, True)}, remaining size {size_to_kb_mb_string(len(self), True)}'

	def __len__(self):
		"""Returns remaining length."""
		return self._full_len - self._start_ptr

	def __enter__(self):
		return self

	def __exit__(self, exception_type, exception_value, traceback):
		self.close()

	@property
	def full_len(self) -> int:
		"""Returns original full length."""
		return self._full_len

	@property
	def binary(self) -> bool:
		"""Returns true, if the data provided is binary."""
		return self._binary

	def read(self, chunk_size: int = None) -> AnyStr:
		"""Read chunk from the data and moves the data pointer behind it.
		If the remaining length is smaller than the chunk_size, the method returns the remaining length only.
		:param chunk_size: chunk to read. If not set, the method reads the entire data."""
		assert self._data is not None, 'StreamReader buffer is invalid. You have probably closed it already.'
		chunk_size = len(self) if chunk_size is None else chunk_size
		chunk_size = min(chunk_size, len(self))
		if chunk_size < 0:
			raise ValueError(f'Chunk size can not be negative number: {chunk_size}')
		self._read_len += chunk_size
		if self._source == Type.Variable:
			self._start_ptr += chunk_size
			return self._data[self._start_ptr - chunk_size: self._start_ptr]
		elif self._source == Type.File:
			self._start_ptr += chunk_size
			return self._data.read(chunk_size)

	@property
	def read_len(self) -> int:
		"""Returns number of bytes read from the stream since its creation."""
		return self._read_len

	def read_as_binary(self, encoding: str, chunk_size: int = None) -> bytes:
		"""Same as read(), but always returns the data in binary format.
		Practically works exactly as read() for binary streams.
		For string streams, the method converts the returned data using the provided encoding to bytes()."""
		if self._binary:
			return self.read(chunk_size)
		else:
			return self.read(chunk_size).encode(encoding)

	def close(self):
		"""Closes the StreamReader. You can not use its instance afterwards."""
		if self._source == Type.File and self._data:
			self._data.close()
		self._data = None



"""VISA communication interface for SCPI-based instrument remote control.
    :version: 1.41.0.89
    :copyright: 2020 by Rohde & Schwarz GMBH & Co. KG
    :license: MIT, see LICENSE for more details.
"""


__version__ = '1.41.0.89'

# Main class
from RsInstrument.RsInstrument import RsInstrument

# Bin data format
from RsInstrument.RsInstrument import BinFloatFormat, BinIntFormat

# Exceptions
from RsInstrument.Internal.InstrumentErrors import RsInstrException, TimeoutException, StatusException, UnexpectedResponseException, ResourceError, DriverValueError

# Callback Event Argument prototypes
from RsInstrument.Internal.IoTransferEventArgs import IoTransferEventArgs

# Logging Mode
from RsInstrument.Internal.ScpiLogger import LoggingMode



"""Utilities extending the driver for methods provided with RsInstrument."""

from typing import List, Tuple
import threading

from ..Internal import Core
from ..Internal import Conversions as Conv
from ..Internal.ScpiLogger import ScpiLogger


class Utilities:
	"""Common utility class.
	Utility functions common for all types of drivers. \n
	Access snippet: ``utils = [DRIVER_PREFIX].utilities``"""
	def __init__(self, core: Core):
		self._core = core

	@property
	def logger(self) -> ScpiLogger:
		"""Scpi Logger interface, see :ref:`here <Logger>` \n
		Access snippet: ``logger = [DRIVER_PREFIX].utilities.logger``"""
		return self._core.io.logger

	@property
	def driver_version(self) -> str:
		"""Returns the instrument driver version."""
		return self._core.driver_version

	@property
	def idn_string(self) -> str:
		"""Returns instrument's identification string - the response on the SCPI command *IDN?"""
		return self._core.io.idn_string

	@property
	def manufacturer(self) -> str:
		"""Returns manufacturer of the instrument."""
		return self._core.io.manufacturer

	@property
	def full_instrument_model_name(self) -> str:
		"""Returns the current instrument's full name e.g. 'FSW26'."""
		return self._core.io.full_model_name

	@property
	def instrument_model_name(self) -> str:
		"""Returns the current instrument's family name e.g. 'FSW'."""
		return self._core.io.model

	@property
	def supported_models(self) -> List[str]:
		"""Returns a list of the instrument models supported by this instrument driver."""
		return self._core.supported_instr_models

	@property
	def instrument_firmware_version(self) -> str:
		"""Returns instrument's firmware version."""
		return self._core.io.firmware_version

	@property
	def instrument_serial_number(self) -> str:
		"""Returns instrument's serial_number."""
		return self._core.io.serial_number

	def query_opc(self, timeout: int = 0) -> int:
		"""SCPI command: *OPC?
		Queries the instrument's OPC bit and hence it waits until the instrument reports operation complete.
		If you define timeout > 0, the VISA timeout is set to that value just for this method call."""
		return self._core.io.query_opc(timeout)

	@property
	def instrument_status_checking(self) -> bool:
		"""Sets / returns Instrument Status Checking.
		When True (default is True), all the driver methods and properties are sending "SYSTem:ERRor?"
		at the end to immediately react on error that might have occurred.
		We recommend to keep the state checking ON all the time. Switch it OFF only in rare cases when you require maximum speed.
		The default state after initializing the session is ON."""
		return self._core.io.query_instr_status

	@instrument_status_checking.setter
	def instrument_status_checking(self, value) -> None:
		"""Sets / returns Instrument Status Checking.
		When True (default is True), all the driver methods and properties are sending "SYSTem:ERRor?"
		at the end to immediately react on error that might have occurred.
		We recommend to keep the state checking ON all the time. Switch it OFF only in rare cases when you require maximum speed.
		The default state after initializing the session is ON."""
		self._core.io.query_instr_status = value

	@property
	def encoding(self) -> str:
		"""Returns string<=>bytes encoding of the session."""
		return self._core.io.encoding

	@encoding.setter
	def encoding(self, value: str) -> None:
		"""Sets string<=>bytes encoding of the session."""
		self._core.io.encoding = value

	@property
	def opc_query_after_write(self) -> bool:
		"""Sets / returns Instrument *OPC? query sending after each command write.
		When True, (default is False) the driver sends *OPC? every time a write command is performed.
		Use this if you want to make sure your sequence is performed command-after-command."""
		return self._core.io.opc_query_after_write

	@opc_query_after_write.setter
	def opc_query_after_write(self, value) -> None:
		"""Sets / returns Instrument *OPC? query sending after each command write.
		When True, (default is False) the driver sends *OPC? every time a write command is performed.
		Use this if you want to make sure your sequence is performed command-after-command."""
		self._core.io.opc_query_after_write = value

	@property
	def bin_float_numbers_format(self) -> Conv.BinFloatFormat:
		"""Sets / returns format of float numbers when transferred as binary data."""
		return self._core.io.bin_float_numbers_format

	@bin_float_numbers_format.setter
	def bin_float_numbers_format(self, value: Conv.BinFloatFormat) -> None:
		"""Sets / returns format of float numbers when transferred as binary data."""
		self._core.io.bin_float_numbers_format = value

	@property
	def bin_int_numbers_format(self) -> Conv.BinIntFormat:
		"""Sets / returns format of integer numbers when transferred as binary data."""
		return self._core.io.bin_int_numbers_format

	@bin_int_numbers_format.setter
	def bin_int_numbers_format(self, value: Conv.BinIntFormat) -> None:
		"""Sets / returns format of integer numbers when transferred as binary data."""
		self._core.io.bin_int_numbers_format = value

	def clear_status(self) -> None:
		"""Clears instrument's status system, the session's I/O buffers and the instrument's error queue."""
		return self._core.io.clear_status()

	def query_all_errors(self) -> List[str] or None:
		"""Queries and clears all the errors from the instrument's error queue.
		The method returns list of strings as error messages. If no error is detected, the return value is None.
		The process is: querying 'SYSTem:ERRor?' in a loop until the error queue is empty.
		If you want to include the error codes, call the query_all_errors_with_codes()"""
		return self._core.io.query_all_syst_errors(include_codes=False)

	def query_all_errors_with_codes(self) -> List[Tuple[int, str]] or None:
		"""Queries and clears all the errors from the instrument's error queue.
		The method returns list of tuples (code: int, message: str). If no error is detected, the return value is None.
		The process is: querying 'SYSTem:ERRor?' in a loop until the error queue is empty."""
		return self._core.io.query_all_syst_errors(include_codes=True)

	@property
	def instrument_options(self) -> List[str]:
		"""Returns all the instrument options.
		The options are sorted in the ascending order starting with K-options and continuing with B-options."""
		return self._core.io.instr_options.get_all()

	def reset(self) -> None:
		"""SCPI command: *RST
		Sends *RST command + calls the clear_status()."""
		self._core.io.reset()
		self.default_instrument_setup()

	def default_instrument_setup(self) -> None:
		"""Custom steps performed at the init and at the reset()."""

	def self_test(self, timeout: int = None) -> Tuple[int, str]:
		"""SCPI command: *TST?
		Performs instrument's self-test.
		Returns tuple (code:int, message: str). Code 0 means the self-test passed.
		You can define the custom timeout in milliseconds. If you do not define it, the default selftest timeout is used (usually 60 secs)."""
		return self._core.io.self_test(timeout)

	def is_connection_active(self) -> bool:
		"""Returns true, if the VISA connection is active and the communication with the instrument still works."""
		return self._core.io.is_connection_active()

	def reconnect(self, force_close: bool = False) -> bool:
		"""If the connection is not active, the method tries to reconnect to the device
		If the connection is active, and force_close is False, the method does nothing.
		If the connection is active, and force_close is True, the method closes, and opens the session again.
		Returns True, if the reconnection has been performed."""
		return self._core.io.reconnect(force_close)

	@property
	def resource_name(self) -> int:
		"""Returns the resource name used in the constructor"""
		return self._core.io.resource_name

	@property
	def opc_timeout(self) -> int:
		"""Sets / returns timeout in milliseconds for all the operations that use OPC synchronization."""
		return self._core.io.opc_timeout

	@opc_timeout.setter
	def opc_timeout(self, value: int) -> None:
		"""Sets / returns timeout in milliseconds for all the operations that use OPC synchronization."""
		self._core.io.opc_timeout = value

	@property
	def visa_timeout(self) -> int:
		"""Sets / returns visa IO timeout in milliseconds."""
		return self._core.io.visa_timeout

	@visa_timeout.setter
	def visa_timeout(self, value) -> None:
		"""Sets / returns visa IO timeout in milliseconds."""
		self._core.io.visa_timeout = value

	@property
	def data_chunk_size(self) -> int:
		"""Sets / returns the maximum size of one block transferred during write/read operations"""
		return self._core.io.data_chunk_size

	@data_chunk_size.setter
	def data_chunk_size(self, chunk_size: int) -> None:
		"""Sets / returns the maximum size of one block transferred during write/read operations."""
		self._core.io.data_chunk_size = chunk_size

	@property
	def visa_manufacturer(self) -> int:
		"""Returns the manufacturer of the current VISA session."""
		return self._core.io.visa_manufacturer

	def process_all_commands(self) -> None:
		"""SCPI command: *WAI
		Stops further commands processing until all commands sent before *WAI have been executed."""
		return self._core.io.write('*WAI')

	def write_str(self, cmd: str) -> None:
		"""Writes the command to the instrument."""
		self._core.io.write(cmd)

	def write(self, cmd: str) -> None:
		"""This method is an alias to the write_str().
		Writes the command to the instrument as string."""
		self._core.io.write(cmd)

	def write_int(self, cmd: str, param: int) -> None:
		"""Writes the command to the instrument followed by the integer parameter:
		e.g.: cmd = 'SELECT:INPUT' param = '2', result command = 'SELECT:INPUT 2'"""
		self._core.io.write(f'{cmd} {param}')

	def write_int_with_opc(self, cmd: str, param: int, timeout: int = None) -> None:
		"""Writes the command with OPC to the instrument followed by the integer parameter:
		e.g.: cmd = 'SELECT:INPUT' param = '2', result command = 'SELECT:INPUT 2'
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(f'{cmd} {param}', timeout)

	def write_float(self, cmd: str, param: float) -> None:
		"""Writes the command to the instrument followed by the boolean parameter:
		e.g.: cmd = 'CENTER:FREQ' param = '10E6', result command = 'CENTER:FREQ 10E6'"""
		self._core.io.write(f'{cmd} {Conv.float_to_str(param)}')

	def write_float_with_opc(self, cmd: str, param: float, timeout: int = None) -> None:
		"""Writes the command with OPC to the instrument followed by the boolean parameter:
		e.g.: cmd = 'CENTER:FREQ' param = '10E6', result command = 'CENTER:FREQ 10E6'
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(f'{cmd} {Conv.float_to_str(param)}', timeout)

	def write_bool(self, cmd: str, param: bool) -> None:
		"""Writes the command to the instrument followed by the boolean parameter:
		e.g.: cmd = 'OUTPUT' param = 'True', result command = 'OUTPUT ON'"""
		self._core.io.write(f'{cmd} {Conv.bool_to_str(param)}')

	def write_bool_with_opc(self, cmd: str, param: bool, timeout: int = None) -> None:
		"""Writes the command with OPC to the instrument followed by the boolean parameter:
		e.g.: cmd = 'OUTPUT' param = 'True', result command = 'OUTPUT ON'
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(f'{cmd} {Conv.bool_to_str(param)}', timeout)

	def query_str(self, query: str) -> str:
		"""Sends the query to the instrument and returns the response as string.
		The response is trimmed of any trailing LF characters and has no length limit."""
		return self._core.io.query_str(query)

	def query(self, query: str) -> str:
		"""This method is an alias to the query_str().
		Sends the query to the instrument and returns the response as string.
		The response is trimmed of any trailing LF characters and has no length limit."""
		return self._core.io.query_str(query)

	def query_bool(self, query: str) -> bool:
		"""Sends the query to the instrument and returns the response as boolean."""
		return self._core.io.query_bool(query)

	def query_int(self, query: str) -> int:
		"""Sends the query to the instrument and returns the response as integer."""
		return self._core.io.query_int(query)

	def query_float(self, query: str) -> float:
		"""Sends the query to the instrument and returns the response as float."""
		return self._core.io.query_float(query)

	def write_str_with_opc(self, cmd: str, timeout: int = None) -> None:
		"""Writes the opc-synced command to the instrument.
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(cmd, timeout)

	def write_with_opc(self, cmd: str, timeout: int = None) -> None:
		"""This method is an alias to the write_str_with_opc().
		Writes the opc-synced command to the instrument.
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(cmd, timeout)

	def query_str_with_opc(self, query: str, timeout: int = None) -> str:
		"""Sends the opc-synced query to the instrument and returns the response as string.
		The response is trimmed of any trailing LF characters and has no length limit.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_str_with_opc(query, timeout)

	def query_with_opc(self, query: str, timeout: int = None) -> str:
		"""This method is an alias to the query_str_with_opc().
		Sends the opc-synced query to the instrument and returns the response as string.
		The response is trimmed of any trailing LF characters and has no length limit.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_str_with_opc(query, timeout)

	def query_bool_with_opc(self, query: str, timeout: int = None) -> bool:
		"""Sends the opc-synced query to the instrument and returns the response as boolean.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_bool_with_opc(query, timeout)

	def query_int_with_opc(self, query: str, timeout: int = None) -> int:
		"""Sends the opc-synced query to the instrument and returns the response as integer.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_int_with_opc(query, timeout)

	def query_float_with_opc(self, query: str, timeout: int = None) -> float:
		"""Sends the opc-synced query to the instrument and returns the response as float.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_float_with_opc(query, timeout)

	def write_bin_block(self, cmd: str, payload: bytes) -> None:
		"""Writes all the payload as binary data block to the instrument.
		The binary data header is added at the beginning of the transmission automatically, do not include it in the payload!!!"""
		self._core.io.write_bin_block(cmd, payload)

	def query_bin_block(self, query: str) -> bytes:
		"""Queries binary data block to bytes.
		Throws an exception if the returned data was not a binary data.
		Returns data:bytes"""
		return self._core.io.query_bin_block(query)

	def query_bin_block_with_opc(self, query: str, timeout: int = None) -> bytes:
		"""Sends a OPC-synced query and returns binary data block to bytes.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_bin_block_with_opc(query, timeout)

	def query_bin_or_ascii_float_list(self, query: str) -> List[float]:
		"""Queries a list of floating-point numbers that can be returned in ASCII format or in binary format.
		- For ASCII format, the list numbers are decoded as comma-separated values.
		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32)."""
		return self._core.io.query_bin_or_ascii_float_list(query)

	def query_bin_or_ascii_float_list_with_opc(self, query: str, timeout: int = None) -> List[float]:
		"""Sends a OPC-synced query and reads a list of floating-point numbers that can be returned in ASCII format or in binary format.
		- For ASCII format, the list numbers are decoded as comma-separated values.
		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32).
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_bin_or_ascii_float_list_with_opc(query, timeout)

	def query_bin_or_ascii_int_list(self, query: str) -> List[int]:
		"""Queries a list of floating-point numbers that can be returned in ASCII format or in binary format.
		- For ASCII format, the list numbers are decoded as comma-separated values.
		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32)."""
		return self._core.io.query_bin_or_ascii_int_list(query)

	def query_bin_or_ascii_int_list_with_opc(self, query: str, timeout: int = None) -> List[int]:
		"""Sends a OPC-synced query and reads a list of floating-point numbers that can be returned in ASCII format or in binary format.
		- For ASCII format, the list numbers are decoded as comma-separated values.
		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32).
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_bin_or_ascii_int_list_with_opc(query, timeout)

	# Write / Read to file
	def query_bin_block_to_file(self, query: str, file_path: str, append: bool = False) -> None:
		"""Queries binary data block to the provided file.
		If append is False, any existing file content is discarded.
		If append is True, the new content is added to the end of the existing file, or if the file does not exit, it is created.
		Throws an exception if the returned data was not a binary data.
		Example for transferring a file from Instrument -> PC:
		query = f"MMEM:DATA? '{INSTR_FILE_PATH}'".
		Alternatively, use the dedicated methods for this purpose:
			- ``send_file_from_pc_to_instrument()``
			- ``read_file_from_instrument_to_pc()``"""
		self._core.io.query_bin_block_to_file(query, file_path, append)

	def query_bin_block_to_file_with_opc(self, query: str, file_path: str, append: bool = False, timeout: int = None) -> None:
		"""Sends a OPC-synced query and writes the returned data to the provided file.
		If append is False, any existing file content is discarded.
		If append is True, the new content is added to the end of the existing file, or if the file does not exit, it is created.
		Throws an exception if the returned data was not a binary data."""
		self._core.io.query_bin_block_to_file_with_opc(query, file_path, append, timeout)

	def write_bin_block_from_file(self, cmd: str, file_path: str) -> None:
		"""Writes data from the file as binary data block to the instrument using the provided command.
		Example for transferring a file from PC -> Instrument:
		cmd = f"MMEM:DATA '{INSTR_FILE_PATH}',".
		Alternatively, use the dedicated methods for this purpose:
			- ``send_file_from_pc_to_instrument()``
			- ``read_file_from_instrument_to_pc()``"""
		self._core.io.write_bin_block_from_file(cmd, file_path)

	def send_file_from_pc_to_instrument(self, source_pc_file: str, target_instr_file: str) -> None:
		"""SCPI Command: MMEM:DATA \n
		Sends file from PC to the instrument"""
		self._core.io.send_file_from_pc_to_instrument(source_pc_file, target_instr_file)

	def read_file_from_instrument_to_pc(self, source_instr_file: str, target_pc_file: str, append_to_pc_file: bool = False) -> None:
		"""SCPI Command: MMEM:DATA? \n
		Reads file from instrument to the PC. \n
		Set the ``append_to_pc_file`` to True if you want to append the read content to the end of the existing PC file"""
		self._core.io.read_file_from_instrument_to_pc(source_instr_file, target_pc_file, append_to_pc_file)

	def get_last_sent_cmd(self) -> str:
		"""Returns the last commands sent to the instrument. Only works in simulation mode"""
		return self._core.get_last_sent_cmd()

	def get_lock(self) -> threading.RLock:
		"""Returns the thread lock for the current session. \n
		By default:
			- If you create standard new [DRIVER_PREFIX] instance with new VISA session, the session gets a new thread lock. You can assign it to other [DRIVER_PREFIX] sessions in order to share one physical instrument with a multi-thread access.
			- If you create new [DRIVER_PREFIX] from an existing session, the thread lock is shared automatically making both instances multi-thread safe.
		You can always assign new thread lock by calling ``driver.utilities.assign_lock()``"""
		return self._core.io.get_lock()

	def assign_lock(self, lock: threading.RLock) -> None:
		"""Assigns the provided thread lock."""
		self._core.io.assign_lock(lock)

	def clear_lock(self):
		"""Clears the existing thread lock, making the current session thread-independent from others that might share the current thread lock."""
		self._core.io.clear_lock()



"""Setup file for creating the RsInstrument package."""

import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.rst").read_text()

# This call to setup() does all the work
setup(
	name="RsInstrument",
	version="1.41.0.89",
	description="VISA or Socket Communication Module for Rohde & Schwarz Instruments",
	long_description=README,
	long_description_content_type="text/x-rst",
	author="Rohde & Schwarz GmbH & Co. KG",
	copyright="Copyright © Rohde & Schwarz GmbH & Co. KG 2020",
	url="https://github.com/Rohde-Schwarz/RsInstrument",
	license="MIT",
	classifiers=[
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python :: 3.6",
	],
	packages=(find_packages(include=['RsInstrument', 'RsInstrument.*'])),
	install_requires=["PyVisa"]
)



"""See the class docstring."""

import re
from typing import Callable

from . import InstrumentOptions as Options, Conversions as Conv
from .ArgSingle import ArgSingle
from .ArgSingleList import ArgSingleList
from .Conversions import BinFloatFormat, BinIntFormat
from .Instrument import Instrument
from .InstrumentSettings import InstrViClearMode, InstrumentSettings, WaitForOpcMode
from .Utilities import parse_token_to_key_and_value, trim_str_response
from .ScpiLogger import LoggingMode
from .InstrumentErrors import RsInstrException


class Core(object):
	"""Main driver component. Provides: \n
		- Main core constructor
		- 'io' interface for all the write / query operations
		- Command parameters string composer for single arguments...
		- Link handlers adding / changing / deleting

		Version history:

		1.40.0 (23.06.2022)
			- Added relative timestamp to the logger.
			- ScpiLogger can read GlobalData class variables making it possible to define common target and reference timestamp for all instances.
			- Logger stream entries are by default immediately flushed, making sure that the log is complete.
			- Added time statistic methods get_total_execution_time(), get_total_time(), reset_time_statistics().

		1.24.0 (03.06.2022)
			- Changed parsing of SYST:ERR? response to tolerate +0,"No Error" response.
			- Added settings integer token OpenTimeout. Example: OpenTimeout=5000. Default: 0
			- Added settings boolean token ExclusiveLock. Example: ExclusiveLock=True. Default: False

		1.23.0 (24.05.2022)
			- Added stripping of trailing commas when parsing the *IDN? response.
			- If the Resource Manager does not find any default VISA implementation, it falls back to R&S VISA - relevant for LINUX or macOS
			- Other typos and formatting corrections.

		1.22.0 (20.04.2022)
			- Added optional parameter timeout to reset()
			- Added query list methods:  query_bool_list, query_bool_list_with_opc

		1.21.0 (07.01.2022)
			- Added logging to UDP port (49200) to integrate with new R&S Instrument Control plugin for Pycharm

		1.20.0 (19.11.2021)
			- Fixed logging strings when device name was a substring of the resource name

		1.18.0 (build 64) 05.11.2021
			- Added setting profile for non-standard instruments. Example of the options string: options='Profile=hm8123'

		1.17.0 (build 63) 15.10.2021
			- Added correct conversion of strings with SI suffixes (e.g.: MHz, KHz, THz, GHz, ms, us) to float and integer

		1.16.0 (build 62) 31.08.2021
			- Changed default encoding of string<=>bin from utf-8 to charmap.
			- Added settable encoding for the session. Property: RsInstrument.encoding

		1.15.0 (build 61) 17.08.2021
			- Added support for EnumExt and EnumExtList
			- Added support for custom scpi enums
			- Improved exception handling in cases where the instrument session is closed.
			- Fixed warning in Instrument.py
			- Fixed Instrument.query_bin_block() for timeout errors
			- Repeated capabilities are now allowed to be integer numbers as well

		1.14.0 (build 53) 12.07.2021
			- Scpi logger time entries now support not only datetime tuples, but also float timestamps
			- changed handling of the syst:err? responses - now they are always Tuple (code, message)
			- StatusException has new field errors_list: List[ Tuple[code, message] ]
			- Added logger.log_status_check_ok property. This allows for skipping lines with 'Status check: OK'

		1.12.0 (build 50) 26.06.2021
			- Added SCPI Logger
			- Simplified constructor's options string format - removed DriverSetup=() syntax:
			Instead of "DriverSetup=(TerminationCharacter='\n')", you use "TerminationCharacter='\n'"
			The original format is still supported.

			- Fixed calling SYST:ERR? even if *STB? returned 0
			- Replaced @ni backend with @ivi for resource manager - this is necessary for the future pyvisa version 1.12+

		1.11.0 (build 49) 09.06.2021
			- Added is_connection_active() + reconnect()

		1.10.1 (build 47) 01.06.2021
			- Fixed bug with error checking when events are defined

		1.10.0 (build 46) 03.05.2021
			- Added methods to Instrument: query_struct_with_opc(), query_str_suppressed_with_opc()

		1.9.0 (build 45) 13.04.2021
			- Added option to set callbacks before_write and before_query
			- When a RepCap has a member with integer number 0 defined, the command string interpretation of such member is '0', not empty string

		1.8.0 (build 43) 19.01.2021
			- Added matching of Enum instrument responses also in short/long form

		1.7.7 (build 42) 26.11.2020
			- Extended ArgSingleList.compose_cmd_string() to 9 arguments

		1.7.6 (build 41) 23.11.2020
			- Extended data types for IntegerExt, FloatExt, IntegerExtArray, FloatExtArray

		1.7.5 (build 40) 12.11.2020
			- Extended 'Conversions' method str_to_str_list() by parameter 'clear_one_empty_item' with default value False

		1.7.4 (build 39) 11.09.2020
			- Fixed parsing of the instrument errors when an error message contains two double quotes

		1.7.3 (build 38) 21.10.2020
			- Added 'UND' to the list of float numbers that are represented as NaN

		1.7.2 (build 37) 10.10.2020
			- SCPI response string conversion to scalar enum: if the string contains ',', the content after it inclusive the comma is ignored

		1.7.1 (build 36) 08.10.2020
			- Fixed Python 3.8.5+ warnings

		1.7.0 (build 34) 30.09.2020
			- Added option to set the termination characters for reading and writing. Until now, it was fixed to '\n' (Linefeed)
			- Replaced 'import visa' with 'import pyvisa' to remove Python 3.8 pyvisa warnings

		Version History:
		1.6.0 (build 33) 17.09.2020
			- Added special characters encoding/decoding in enums

		1.4.0 (build 32) 17.09.2020
			- Added recognition of RsVisa library location for linux when using options string 'SelectVisa=rs'
			- Fixed bug in reading binary data 16 bit

		1.3.0 (build 31) 04.09.2020
			- added DRIVERSETUP_QUERYOPT to the driver's option string
			- *OPT? is no longer performed at the init, but only at the first access to the options string.
				In addition, the *OPT? query is executed with 1000 ms timeout, and the errors are suppressed

		1.2.0 (build 30), 03.08.2020
			- Fixed NRP-Z session parameters: vxi_capable = False, io_segment_size = 1000000

		1.1.0 (build 29), 20.06.2020
			- Added RepeatedCapability and base class CommandsGroup
			- Fixed simulation mode switching

		0.9.3 (build 25), 23.04.2020 - Fixed composition of optional arguments in ArgSingleList and ArgSingle
		0.9.2 (build 24), 13.11.2019 - Added recognition of special values for enum return strings
		0.9.1 - Added read / write to file, refactored internals to work with streams
		0.9.0 - First Version created."""

	driver_version: str = ''
	"""Placeholder for the driver version string."""

	def __init__(
			self,
			resource_name: str,
			id_query: bool = True,
			reset: bool = False,
			driver_options: str = None,
			user_options: str = None,
			direct_session: object = None):
		"""Initializes new driver session. For cleaner code, use the class methods: \n
		- Core.from_existing_session() - initializes a new Core with an existing pyvisa session."""

		self.core_version = '1.40.0'
		self.simulating = False
		self.supported_idn_patterns = []
		self.supported_instr_models = []
		self.resource_name = resource_name

		self._args_single_list = ArgSingleList()
		self.sett_dr = self._parse_init_settings_string(driver_options)
		self._apply_settings_to_core(self.sett_dr)
		self.sett_user = self._parse_init_settings_string(user_options)
		self._apply_settings_to_core(self.sett_user)

		# Typical settings for the Core
		self._instrumentSettings = InstrumentSettings(
			InstrViClearMode.execute_on_all,  # Instrument viClear mode
			False,  # Full model name. True: SMW200A, False: SMW
			0,  # Delay by each write
			0,  # Delay by each read
			100000,  # Max chunk read / write size in bytes
			WaitForOpcMode.stb_poll,  # Waiting for OPC Mode: Status byte polling
			30000,  # OPC timeout
			10000,  # VISA timeout
			60000,  # Self-test timeout
			Options.ParseMode.Auto,  # *OPT? response parsing mode
			BinFloatFormat.Single_4bytes,  # Format for parsing of binary float numbers
			BinIntFormat.Integer32_4bytes,  # Format for parsing of binary integer numbers
			False,  # OPC query after each setting
			LoggingMode.Off  # Logging mode
		)

		self._instrumentSettings.apply_option_settings(self.sett_dr)
		self._instrumentSettings.apply_option_settings(self.sett_user)

		handle = self._resolve_direct_session(direct_session)
		self.io = Instrument(self.resource_name, self.simulating, self._instrumentSettings, handle)
		self.io.query_instr_status = True
		# Update the resource name if it changed, for example because of the direct session
		self.resource_name = self.io.resource_name
		self.allow_reconnect = self.io.allow_reconnect

		self._apply_settings_to_instrument(self._instrumentSettings)
		self.io.set_simulating_cmds()

		if id_query:
			self.io.fits_idn_pattern(self.supported_idn_patterns, self.supported_instr_models)

		if reset:
			self.io.reset()
		else:
			self.io.check_status()

	@classmethod
	def from_existing_session(cls, session: object, driver_options: str = None) -> 'Core':
		"""Creates a new Core object with the entered 'session' reused."""
		# noinspection PyTypeChecker
		return cls(resource_name=None, id_query=False, reset=False, driver_options=driver_options, user_options=None, direct_session=session)

	def __str__(self):
		return f"Core session '{self.io.resource_name}'"

	def _resolve_direct_session(self, direct_session):
		# Resolve the direct_session to handle. Options for direct_session type:
		# - VisaSession object, retrieved from the driver's RsInstrument.get_session_handle() method
		# - string in case of a simulation session
		handle = direct_session
		if not direct_session:
			return None
		# Check if the entered 'direct_session' is either the driver object or the Visa session
		if hasattr(direct_session, 'get_session_handle'):
			if not hasattr(direct_session, '_core'):
				raise RsInstrException('Direct session is a class type. It must be an instance of the top-level driver class.')
			handle = direct_session.get_session_handle()
		# If the handle is a simulating session, change the session to simulating and set disable the 'from existing session' feature
		if isinstance(handle, str):
			mand_string = 'Simulating session, resource name '
			if mand_string in handle:
				self.resource_name = handle[len(mand_string):].strip().strip("'").strip()
				self.simulating = True
				handle = None
		return handle

	def set_link_handler(self, link_name: str, handler: Callable) -> Callable:
		"""Adds / Updates link handler for the entered link_name.
		Handler API: handler(event_args: ArgLinkedEventArgs)
		Returns the previous registered handler, or None if no handler was registered before."""
		return self.io.set_link_handler(link_name, handler)

	def del_link_handler(self, link_name: str) -> Callable:
		"""Deletes link handler for the link_name.
		Returns the deleted handler, or None if none existed."""
		return self.io.del_link_handler(link_name)

	def del_all_link_handlers(self) -> int:
		"""Deletes all the link handlers.
		Returns number of deleted links."""
		return self.io.del_all_link_handlers()

	# noinspection PyMethodMayBeStatic
	def _parse_init_settings_string(self, text: str) -> dict:
		"""Parses init string to a dictionary of settings: name -> value."""
		tokens = {}
		if not text:
			return tokens

		# Text enclosed in single brackets '' must have the commas escaped
		literal_pattern = r"'([^']+)'"
		while True:
			# literal loop
			m = re.search(literal_pattern, text)
			if not m:
				break
			lit_part = '"' + m.group(1).replace(',', '<COMMA_ESC>') + '"'
			text = text.replace(m.group(0), lit_part)

		# Remove all the class-options enclosed by round brackets e.g. "<groupName>=(<groupTokens>)"
		group_pattern = r'(\w+)\s*=\s*\(([^\)]*)\)'
		# Match class-settings, add them as separate keys with groupName_Key
		while True:
			# Group loop
			m = re.search(group_pattern, text)
			if not m:
				break
			text = text.replace(m.group(0), '')
			group_name = m.group(1).upper()
			group_tokens = m.group(2).strip().split(',')
			for token in group_tokens:
				key, value = parse_token_to_key_and_value(token)
				if value:
					tokens[f'{group_name}_{key.upper()}'] = value

		# All groups are removed from the text, now we can use splitting on commas and remove white-space-only elements
		for token in text.split(','):
			key, value = parse_token_to_key_and_value(token.replace('<COMMA_ESC>', ','))
			if value:
				tokens[key.upper()] = value
		return tokens

	def _apply_settings_to_core(self, settings: dict) -> None:
		"""Applies settings relevant for the Core from the dictionary."""
		value = settings.get('SIMULATE')
		if value:
			self.simulating = Conv.str_to_bool(value)

		value = settings.get('SUPPORTEDINSTRMODELS')
		if value:
			self.supported_instr_models = [*map(trim_str_response, value.split('/'))]

		value = settings.get('SUPPORTEDIDNPATTERNS')
		if value:
			self.supported_idn_patterns = [*map(trim_str_response, value.split('/'))]

	def _apply_settings_to_instrument(self, settings: InstrumentSettings) -> None:
		"""Applies settings relevant for the Instrument from the InstrumentSettings structure."""
		if settings.instrument_status_check is not None:
			self.io.query_instr_status = settings.instrument_status_check
		if self.simulating and settings.instrument_simulation_idn_string is not None:
			self.io.idn_string = settings.instrument_simulation_idn_string

	def compose_cmd_arg_param(
			self, arg1: ArgSingle, arg2: ArgSingle = None, arg3: ArgSingle = None, arg4: ArgSingle = None, arg5: ArgSingle = None, arg6: ArgSingle = None) -> str:
		"""Composes command parameter string based on the single argument definition."""
		return self._args_single_list.compose_cmd_string(arg1, arg2, arg3, arg4, arg5, arg6)

	def get_last_sent_cmd(self) -> str:
		"""Returns the last commands sent to the instrument. Only works in simulation mode"""
		return self.io.get_last_sent_cmd()

	def get_session_handle(self):
		"""Returns the underlying pyvisa session."""
		return self.io.get_session_handle()

	def close(self):
		"""Closes the Core session."""
		self.io.close()
		self.io = None



"""See the class docstring."""

import re
from enum import Enum

from .Utilities import trim_str_response


class ParseMode(Enum):
	"""Options parse mode enum."""
	Skip = 0
	KeepOriginal = 1
	KeepBeforeDash = 2
	KeepAfterDash = 3
	Auto = 4


class Options(object):
	"""Class for handling the instrument options - parsing from the *OPT? string and providing method get_all()"""
	_optionsList = []

	def __init__(self, options_str: str, mode=ParseMode.Auto):
		"""Initializes the options with the *OPT? return string."""
		self._initialize_from_string(options_str, mode)

	def __str__(self):
		return ','.join(self._optionsList)

	def _initialize_from_string(self, options_str: str, mode: ParseMode):
		"""Fills the self._optionsList from the entered 'options_str'."""
		if mode == ParseMode.Skip:
			return

		parse_patt = r'(.?)(K|B)(\d+)(.*)$'
		options = trim_str_response(options_str).split(',')
		new_opts = []
		for x in options:
			has_dash = '-' in x
			if has_dash:
				dash_ix = x.index('-')
				before = x[0:dash_ix + 1].strip('-')
				after = x[dash_ix:].strip('-')

				if mode is ParseMode.KeepBeforeDash:
					x = before

				elif mode is ParseMode.KeepAfterDash:
					x = after

				elif mode is ParseMode.Auto:
					found_before = re.match(parse_patt, before)
					found_after = re.match(parse_patt, after)
					if found_before is not None and found_after is not None:
						x = after if len(found_after.group(0)) >= len(found_before.group(0)) else before
					elif found_before is not None:
						x = before
					elif found_after is not None:
						x = after

			if len(x) > 0:
				new_opts.append(x)

		# Remove duplicates
		new_opts = set(new_opts)

		# Create a weighting number
		result = []
		for x in new_opts:
			sort_value = x
			m = re.match(parse_patt, x)
			if m:
				kb_weight = '02' if m.group(2) == 'B' else '01'
				sort_value = '{0}{1}{2:09d}{3}'.format(m.group(1), kb_weight, int(m.group(3)), m.group(4))

			result.append('{0} -> {1}'.format(sort_value, x))

		# Sort keys in that dictionary and then reconstruct the original options
		result.sort()
		self._optionsList = [x.split(' -> ')[1] for x in result]

	def get_all(self):
		"""Returns all the options."""
		return self._optionsList



"""See the class docstring."""

from enum import Enum

from .Conversions import list_to_csv_quoted_str, value_to_quoted_str, list_to_csv_str, value_to_str, enum_list_to_str, enum_scalar_to_str, enum_ext_scalar_to_str, enum_ext_list_to_str
from .Types import DataType
from .InstrumentErrors import RsInstrException


def value_to_scpi_string(data, data_type: DataType):
	"""Method to be used in the driver implementation.
	Convert data to SCPI string parameter: data -> str.
	Does not work with enum data types."""
	if data_type.is_list:
		assert isinstance(data, list), f"Expected command parameter list, actual data type: {type(data)}. Value: {data}"
	else:
		assert not isinstance(data, list), f"Expected command parameter scalar, actual data type: {type(data)}. Value: {data}"
	# Strings are enclosed by single quotes
	if data_type == DataType.StringList:
		assert all(isinstance(x, str) for x in data), f"Expected command parameter list of strings, detected one or more elements of non-string type. Value: {data}"
		return list_to_csv_quoted_str(data)
	elif data_type == DataType.String:
		assert isinstance(data, str), f"Expected command parameter string, actual data type: {type(data)}. Value: {data}"
		return value_to_quoted_str(data)

	# Raw string is not enclosed by quotes
	elif data_type == DataType.RawStringList:
		assert all(isinstance(x, str) for x in data), f"Expected command parameter list of strings, detected one or more elements of non-string type. Value: {data}"
		return list_to_csv_str(data)
	elif data_type == DataType.RawString:
		assert isinstance(data, str), f"Expected command parameter string, actual data type: {type(data)}. Value: {data}"
		return value_to_str(data)

	elif data_type == DataType.BooleanList:
		assert all(type(x) == bool for x in data), f"Expected command parameter list of booleans, detected one or more elements of non-boolean type. Value: {data}"
		return list_to_csv_str(data)
	elif data_type == DataType.Boolean:
		assert type(data) == bool, f"Expected command parameter boolean, actual data type: {type(data)}. Value: {data}"
		return value_to_str(data)

	# For integer and float, allow them to be mixed
	elif data_type == DataType.IntegerList or data_type == DataType.FloatList:
		assert all((isinstance(x, int) or isinstance(x, float)) and type(x) != bool for x in data), f"Expected command parameter list of numbers, detected one or more elements of non-number type. Value: {data}"
		return list_to_csv_str(data)
	elif data_type == DataType.Integer or data_type == DataType.Float:
		assert (isinstance(data, int) or isinstance(data, float)) and type(data) != bool, f"Expected command parameter number, actual data type: {type(data)}. Value: {data}"
		return value_to_str(data)

	# For integer and float extended, allow them to be mixed including the boolean type
	elif data_type == DataType.IntegerExtList or data_type == DataType.FloatExtList:
		assert all((isinstance(x, int) or isinstance(x, float) or isinstance(x, bool)) for x in data), f"Expected command parameter list of numbers or booleans, detected one or more elements of non-number type. Value: {data}"
		return list_to_csv_str(data)
	elif data_type == DataType.IntegerExt or data_type == DataType.FloatExt:
		assert (isinstance(data, int) or isinstance(data, float) or isinstance(data, bool)), f"Expected command parameter number or boolean, actual data type: {type(data)}. Value: {data}"
		return value_to_str(data)
	else:
		raise RsInstrException(f"Unsupported data type: '{type(data_type)}'.")


class ConverterToScpiString:
	"""Converter from argument value to SCPI string.
	Provides method get_value(arg_value) -> str
	"""

	def __init__(self, data_type: DataType, enum_type: Enum = None):
		self.enum_type = enum_type
		self.data_type = data_type
		self.element_type = self.data_type.element_type
		if self.element_type == DataType.Enum or self.element_type == DataType.EnumExt:
			assert self.enum_type, f"For data_type {data_type.name}, you have to define the enum_type variable."

	def get_value(self, data) -> str:
		"""Returns SCPI string converted from the argument data."""
		if self.data_type.is_list:
			assert isinstance(data, list), f"Expected command parameter list, actual data type: {type(data)}. Value: {data}"
		else:
			assert not isinstance(data, list), f"Expected command parameter scalar, actual data type: {type(data)}. Value: {data}"

		if self.data_type == DataType.Enum:
			return enum_scalar_to_str(data, self.enum_type)
		if self.data_type == DataType.EnumExt:
			return enum_ext_scalar_to_str(data, self.enum_type)
		if self.data_type == DataType.EnumList:
			return enum_list_to_str(data, self.enum_type)
		if self.data_type == DataType.EnumExtList:
			return enum_ext_list_to_str(data, self.enum_type)

		return value_to_scpi_string(data, self.data_type)
